#! /usr/bin/env python
"""Creation of a profile of studies as part of the data operations"""
from __future__ import annotations

from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.docker_manager
import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class Profile:
    """Create a profile"""

    ci_supported: bool = Field(default=False)

    class ProfileSettings(
        colrev.package_manager.package_settings.DefaultSettings, BaseModel
    ):
        """Profile settings"""

        endpoint: str
        version: str

    settings_class = ProfileSettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.review_manager = data_operation.review_manager
        self.data_operation = data_operation

        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"

        self.settings = self.settings_class(**settings)

        # output_dir = self.review_manager.get_path(Filepaths.OUTPUT_DIR)

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_package = {
            "endpoint": "colrev.profile",
            "version": "0.1",
        }
        operation.review_manager.settings.data.data_package_endpoints.append(
            add_package
        )
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add {operation.type} profile",
        )

    def _update_profile(self, silent_mode: bool) -> None:
        """Create a profile of the sample"""

        self.review_manager.logger.info("Create sample profile")

        def prep_records(*, records: dict) -> pd.DataFrame:
            for record in records.values():
                record["outlet"] = record.get(
                    Fields.JOURNAL, record.get(Fields.BOOKTITLE, "NA")
                )

            records_df = pd.DataFrame.from_records(list(records.values()))

            required_cols = [
                Fields.ID,
                Fields.ENTRYTYPE,
                Fields.AUTHOR,
                Fields.TITLE,
                Fields.JOURNAL,
                Fields.BOOKTITLE,
                "outlet",
                Fields.YEAR,
                Fields.VOLUME,
                Fields.NUMBER,
                Fields.PAGES,
                Fields.DOI,
            ]
            available_cols = records_df.columns.intersection(list(set(required_cols)))
            cols = [x for x in required_cols if x in available_cols]
            records_df = records_df[cols]
            return records_df

        def prep_observations(
            *, prepared_records_df: pd.DataFrame, records: dict
        ) -> pd.DataFrame:
            included_papers = [
                ID
                for ID, record in records.items()
                if record[Fields.STATUS]
                in [
                    RecordState.rev_synthesized,
                    RecordState.rev_included,
                ]
                and record.get(Fields.YEAR, FieldValues.UNKNOWN).isdigit()
            ]
            observations = prepared_records_df[
                prepared_records_df[Fields.ID].isin(included_papers)
            ].copy()
            observations.year = observations.year.astype(int)
            missing_outlet = observations[observations["outlet"].isnull()][
                Fields.ID
            ].tolist()
            if len(missing_outlet) > 0:
                self.review_manager.logger.info(f"No outlet: {missing_outlet}")
            return observations

        # if not status.get_completeness_condition():
        #     self.review_manager.logger.warning(
        #  f"{Colors.RED}Sample not completely processed!{Colors.END}")

        records = self.review_manager.dataset.load_records_dict()

        output_dir = self.review_manager.path / Path("output")
        output_dir.mkdir(exist_ok=True)

        prepared_records_df = prep_records(records=records)  # .values()
        observations = prep_observations(
            prepared_records_df=prepared_records_df, records=records
        )

        if observations.empty:
            self.review_manager.logger.info("No sample/observations available")
            return

        self.review_manager.logger.info("Generate output/sample.csv")
        observations.to_csv(output_dir / Path("sample.csv"), index=False)

        tabulated = pd.pivot_table(
            observations[["outlet", Fields.YEAR]],
            index=["outlet"],
            columns=[Fields.YEAR],
            aggfunc=len,
            fill_value=0,
            margins=True,
        )
        # Fill missing years with 0 columns
        years = range(
            min(e for e in tabulated.columns if isinstance(e, int)),
            max(e for e in tabulated.columns if isinstance(e, int)) + 1,
        )
        for year in years:
            if year not in tabulated.columns:
                tabulated[year] = 0
        year_list = list(years)
        year_list.extend(["All"])  # type: ignore
        tabulated = tabulated[year_list]
        tabulated.sort_values(by=("All"), ascending=True, inplace=True)

        self.review_manager.logger.info("Generate profile output/journals_years.csv")
        tabulated.to_csv(output_dir / Path("journals_years.csv"))

        tabulated = pd.pivot_table(
            observations[[Fields.ENTRYTYPE, Fields.YEAR]],
            index=[Fields.ENTRYTYPE],
            columns=[Fields.YEAR],
            aggfunc=len,
            fill_value=0,
            margins=True,
        )
        self.review_manager.logger.info("Generate output/ENTRYTYPES.csv")
        tabulated.to_csv(output_dir / Path("ENTRYTYPES.csv"))

        self.review_manager.logger.info(f"Files are available in {output_dir.name}")

    def update_data(
        self,
        records: dict,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the data/profile"""

        self._update_profile(silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        # Note : automatically set all to True / synthesized
        for syn_id in list(synthesized_record_status_matrix.keys()):
            synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [profile data endpoint]: "

        advice = {
            "msg": f"{data_endpoint}" + "\n    - The profile is created automatically ",
            "detailed_msg": "TODO",
        }
        return advice



#! /usr/bin/env python
"""CoLRev data operation: extract data, analyze, and synthesize."""
from __future__ import annotations

import typing
from pathlib import Path

import colrev.packages.grobid_tei.src.grobid_tei
import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState


class Data(colrev.process.operation.Operation):
    """Class supporting structured and unstructured
    data extraction, analysis and synthesis"""

    _pad = 0
    type = OperationsType.data

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.package_manager = self.review_manager.get_package_manager()

    def get_record_ids_for_synthesis(self, records: dict) -> list:
        """Get the IDs of records for the synthesis"""
        return [
            ID
            for ID, record in records.items()
            if record[Fields.STATUS]
            in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]
        ]

    def reading_heuristics(self) -> list:
        """Determine heuristics for the reading process"""

        enlit_list = []
        records = self.review_manager.dataset.load_records_dict()
        for relevant_record_id in self.get_record_ids_for_synthesis(records):
            enlit_status = str(records[relevant_record_id][Fields.STATUS])
            enlit_status = enlit_status.replace("rev_included", "").replace(
                "rev_synthesized", "synthesized"
            )
            enlit_list.append(
                {
                    Fields.ID: relevant_record_id,
                    "score": 0,
                    "score_intensity": 0,
                    Fields.STATUS: enlit_status,
                }
            )

        tei_path = colrev.packages.grobid_tei.src.grobid_tei.GROBIDTEI.TEI_PATH_RELATIVE
        required_records_ids = self.get_record_ids_for_synthesis(records)

        missing = []
        for required_records_id in required_records_ids:
            tei_file = tei_path / Path(f"{required_records_id}.tei.xml")
            if not tei_file.is_file():
                missing.append(required_records_id)

            tei_doc = self.review_manager.get_tei(
                tei_path=tei_file,
            )
            tei_doc.mark_references(records=records)
            data = tei_doc.get_tei_str()
            for enlit_item in enlit_list:
                id_string = f'ID="{enlit_item[Fields.ID]}"'
                if id_string in data:
                    enlit_item["score"] += 1
                enlit_item["score_intensity"] += data.count(id_string)

        if len(missing) > 0:
            print(f"Records with missing tei file: {missing}")

        enlit_list = sorted(enlit_list, key=lambda d: d["score"], reverse=True)

        return enlit_list

    def setup_custom_script(self) -> None:
        """Setup a custom data script"""

        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.ops", filename=Path("custom_scripts/custom_data_script.py")
        )
        if filedata:
            with open("custom_data_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(Path("custom_data_script.py"))

        new_data_endpoint = {"endpoint": "custom_data_script"}

        self.review_manager.settings.data.data_package_endpoints.append(
            new_data_endpoint
        )
        self.review_manager.save_settings()

    def _pre_data(self, *, records: dict, silent_mode: bool) -> None:
        if not silent_mode:
            self.review_manager.logger.info("Data")
            self.review_manager.logger.info(
                "The data operation covers different forms of data extraction, "
                "analysis, and synthesis."
            )
            self.review_manager.logger.info(
                "See https://colrev-environment.github.io/colrev/manual/data/data.html"
            )
        self._pad = min((max(len(ID) for ID in list(records.keys()) + [""]) + 2), 35)

    def _get_synthesized_record_status_matrix(self, records: dict) -> dict:
        included = self.get_record_ids_for_synthesis(records)

        # TBD: do we assume that records are not changed by the processes?
        # records = self.review_manager.dataset.load_records_dict()

        # synthesized_record_status_matrix (paper IDs x endpoint):
        # each endpoint sets synthesized = True/False
        # and if a paper has synthesized=True in all fields,
        # its overall status is set to synthesized
        # Some endpoints may always set synthesized
        default_row = {
            df["endpoint"]: False
            for df in self.review_manager.settings.data.data_package_endpoints
        }
        synthesized_record_status_matrix = {ID: default_row.copy() for ID in included}

        # if self.review_manager.verbose_mode:
        #     self.review_manager.p_printer.pprint(synthesized_record_status_matrix)
        return synthesized_record_status_matrix

    def _update_record_status_matrix(
        self, *, records: dict, synthesized_record_status_matrix: dict
    ) -> bool:
        records_changed = False
        for (
            record_id,
            individual_status_dict,
        ) in synthesized_record_status_matrix.items():
            if all(x for x in individual_status_dict.values()):
                if records[record_id][Fields.STATUS] != RecordState.rev_synthesized:
                    if self.review_manager.verbose_mode:
                        self.review_manager.report_logger.info(
                            f" {record_id}".ljust(self._pad, " ")
                            + "set colrev_status to synthesized"
                        )
                        self.review_manager.logger.info(
                            f" {record_id}".ljust(self._pad, " ")
                            + "set colrev_status to synthesized"
                        )

                if RecordState.rev_synthesized != records[record_id][Fields.STATUS]:
                    # pylint: disable=colrev-direct-status-assign
                    records[record_id].update(colrev_status=RecordState.rev_synthesized)
                    records_changed = True
            else:
                if RecordState.rev_included != records[record_id][Fields.STATUS]:
                    # pylint: disable=colrev-direct-status-assign
                    records[record_id].update(colrev_status=RecordState.rev_included)
                    records_changed = True
        return records_changed

    def _post_data(self, *, silent_mode: bool) -> None:
        # if self.review_manager.verbose_mode:
        #     self.review_manager.p_printer.pprint(synthesized_record_status_matrix)

        if self.review_manager.verbose_mode and not silent_mode:
            print()

        if not silent_mode:
            self.review_manager.logger.info(
                f"{Colors.GREEN}Completed data operation{Colors.END}"
            )
        if self.review_manager.in_ci_environment():
            print("\n\n")

    @colrev.process.operation.Operation.decorate()
    def main(
        self,
        *,
        selection_list: typing.Optional[list] = None,
        records: typing.Optional[dict] = None,
        silent_mode: bool = False,
    ) -> dict:
        """Data operation (main entrypoint)

        silent_mode: for review_manager checks
        """

        if not records:
            records = self.review_manager.dataset.load_records_dict()

        self._pre_data(records=records, silent_mode=silent_mode)

        synthesized_record_status_matrix = self._get_synthesized_record_status_matrix(
            records
        )

        for (
            data_package_endpoint
        ) in self.review_manager.settings.data.data_package_endpoints:
            if selection_list:
                if not any(
                    x in data_package_endpoint["endpoint"] for x in selection_list
                ):
                    continue

            if not silent_mode:
                print()
                self.review_manager.logger.info(
                    f"Data: {data_package_endpoint['endpoint'].replace('colrev.', '')}"
                )

            data_class = self.package_manager.get_package_endpoint_class(
                package_type=EndpointType.data,
                package_identifier=data_package_endpoint["endpoint"],
            )

            endpoint = data_class(data_operation=self, settings=data_package_endpoint)

            endpoint.update_data(  # type: ignore
                records, synthesized_record_status_matrix, silent_mode=silent_mode
            )

            endpoint.update_record_status_matrix(  # type: ignore
                synthesized_record_status_matrix,
                data_package_endpoint["endpoint"],
            )

            if self.review_manager.verbose_mode and not silent_mode:
                msg = f"Updated {endpoint.settings.endpoint}"  # type: ignore
                self.review_manager.logger.info(msg)

        records_status_changed = self._update_record_status_matrix(
            records=records,
            synthesized_record_status_matrix=synthesized_record_status_matrix,
        )
        if records_status_changed:
            self.review_manager.dataset.save_records_dict(records)

        self._post_data(silent_mode=silent_mode)

        no_endpoints_registered = 0 == len(
            self.review_manager.settings.data.data_package_endpoints
        )

        return {
            "ask_to_commit": self.review_manager.dataset.has_record_changes(),
            "no_endpoints_registered": no_endpoints_registered,
        }



#! /usr/bin/env python3
"""Command-line interface for CoLRev."""
from __future__ import annotations

import logging
import subprocess  # nosec
import sys
import time
import typing
import webbrowser
from functools import partial
from functools import wraps
from pathlib import Path

import click
import click_repl
import inquirer
import pandas as pd
from git.exc import GitCommandError

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.ops.check
import colrev.package_manager.package_manager
import colrev.record.record
import colrev.review_manager
import colrev.ui_cli.add_package_to_settings
import colrev.ui_cli.cli_status_printer
import colrev.ui_cli.cli_validation
import colrev.ui_cli.dedupe_errors
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import Filepaths
from colrev.constants import RecordState
from colrev.constants import ScreenCriterionType

# pylint: disable=too-many-lines
# pylint: disable=redefined-outer-name
# pylint: disable=too-many-arguments
# pylint: disable=unused-argument
# pylint: disable=superfluous-parens
# pylint: disable=too-many-locals
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-locals
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-positional-arguments

# Note: autocompletion needs bash/... activation:
# https://click.palletsprojects.com/en/7.x/bashcomplete/

EXACT_CALL = "colrev " + subprocess.list2cmdline(sys.argv[1:])  # nosec

PACKAGE_MANAGER = colrev.package_manager.package_manager.PackageManager()
# TYPE_IDENTIFIER_ENDPOINT_DICT = PACKAGE_MANAGER.load_type_identifier_endpoint_dict()

SHELL_MODE = False


def _add_endpoint_interactively(add: str, endpoint_type: EndpointType) -> str:
    """Add package interactively"""
    if add != "add_interactively":
        return add
    packages = PACKAGE_MANAGER.discover_packages(package_type=endpoint_type)
    questions = [
        inquirer.List(
            "package",
            message=f"Select a {endpoint_type.name} package to add using down/up/space:",
            choices=sorted([p["package_endpoint_identifier"] for p in packages]),
        )
    ]
    answers = inquirer.prompt(questions)
    endpoint = answers["package"]

    if not PACKAGE_MANAGER.is_installed(endpoint):
        print(f"{Colors.GREEN}Install package{Colors.END}")
        PACKAGE_MANAGER.install(packages=[endpoint])

    return endpoint


def _select_source_interactively(
    selected: str, review_manager: colrev.review_manager.ReviewManager
) -> str:
    """Select a source interactively"""
    if selected != "select_interactively":
        return selected

    sources = [str(s.filename) for s in review_manager.settings.sources]
    questions = [
        inquirer.Checkbox(
            "source",
            message="Select search source(s) using down/up/space:",
            choices=sorted(sources),
        )
    ]
    answers = inquirer.prompt(questions)
    return ",".join(answers["source"])


def _add_screen_criterion_interactively(
    add: str, screen_operation: colrev.ops.screen.Screen
) -> None:
    """Add screening criterion interactively"""
    if add != "add_interactively":
        assert add.count(",") == 2
        (
            criterion_name,
            criterion_type_str,
            criterion_explanation,
        ) = add.split(",")

    else:

        questions = [
            inquirer.Text(
                "criterion_name", message="Criterion name (e.g., empirical study)"
            ),
            inquirer.Text("criterion_explanation", message="Criterion explanation"),
            inquirer.List(
                "criterion_type",
                message="Criterion type",
                choices=[x.name for x in ScreenCriterionType],
            ),
        ]
        answers = inquirer.prompt(questions)
        criterion_name = answers["criterion_name"]
        criterion_explanation = answers["criterion_explanation"]
        criterion_type_str = answers["criterion_type"]

    criterion_type = ScreenCriterionType[criterion_type_str]
    criterion = colrev.settings.ScreenCriterion(
        explanation=criterion_explanation,
        criterion_type=criterion_type,
        comment="",
    )

    screen_operation.add_criterion(criterion_name=criterion_name, criterion=criterion)


def get_search_files() -> list:
    """Get the search files (for click choices)"""
    # Take the filenames from sources because there may be API searches
    # without files (yet)
    try:
        review_manager = colrev.review_manager.ReviewManager()
        return [str(x.filename) for x in review_manager.settings.sources]
    except Exception:  # pylint: disable=broad-exception-caught
        return []


class SpecialHelpOrder(click.Group):
    """Order for cli commands in help page overview"""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        self.help_priorities: dict = {}
        super().__init__(*args, **kwargs)

    def get_help(self, ctx: click.core.Context):  # type: ignore
        self.list_commands = self.list_commands_for_help  # type: ignore  # noqa
        return super().get_help(ctx)

    def list_commands_for_help(self, ctx: click.core.Context) -> typing.Generator:
        """reorder the list of commands when listing the help"""
        commands = super().list_commands(ctx)
        return (
            c[1]
            for c in sorted(
                (self.help_priorities.get(command, 1), command) for command in commands
            )
        )

    def command(self, *args, **kwargs):  # type: ignore
        """Behaves the same as `click.Group.command()` except capture
        a priority for listing command names in help.
        """
        help_priority = kwargs.pop("help_priority", 1)
        help_priorities = self.help_priorities

        def decorator(fun):  # type: ignore
            # pylint: disable=super-with-arguments
            cmd = super(SpecialHelpOrder, self).command(*args, **kwargs)(fun)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


def catch_exception(func=None, *, handle) -> typing.Any:  # type: ignore
    """Catch typical cli exceptions (e.g., CoLRevException)"""
    if not func:
        return partial(catch_exception, handle=handle)

    # pylint: disable=inconsistent-return-statements
    @wraps(func)
    def wrapper(*args, **kwargs) -> None:  # type: ignore
        try:
            return func(*args, **kwargs)
        except colrev_exceptions.CoLRevException as exc:
            if kwargs.get("verbose", False):
                raise exc
            print(exc)

    return wrapper


@click.group(cls=SpecialHelpOrder)
@click.pass_context
def main(ctx: click.core.Context) -> None:
    """CoLRev commands:

    \b
    status        Shows status, how to proceed, and checks for errors

    \b
    init          Initialize (define review objectives and type)
    retrieve      Search, load, prepare, and deduplicate metadata records
    prescreen     Exclude records based on metadata
    pdfs          Get and prepare PDFs
    screen        Include records based on PDFs
    data          Complete selected forms of data analysis and synthesis

    \b
    validate      Validate changes in the previous commit

    Recommended workflow: colrev status > colrev OPERATION > colrev validate

    Documentation:  https://colrev-environment.github.io/colrev/
    """

    try:
        if ctx.invoked_subcommand == "shell":
            ctx.obj = {"review_manager": colrev.review_manager.ReviewManager()}
    except colrev.exceptions.RepoSetupError:
        pass


def get_review_manager(
    ctx: click.core.Context, review_manager_params: dict
) -> colrev.review_manager.ReviewManager:
    """Get review_manager instance. If it's available in ctx object, reuse that
    if not creates a new one, Once created will update the review_manager with
    the given parameters. If params requires review_manager to be reloaded, will
    reload it
    """

    review_manager_params["exact_call"] = ctx.command_path
    try:
        review_manager = ctx.obj["review_manager"]
        if (
            "navigate_to_home_dir" in review_manager_params
            or "path_str" in review_manager_params
            or "skip_upgrade" in review_manager_params
        ):
            print("init review manager object ...")
            review_manager = colrev.review_manager.ReviewManager(
                **review_manager_params
            )
            ctx.obj["review_manager"] = review_manager
        else:
            review_manager.update_config(**review_manager_params)
        if SHELL_MODE:
            review_manager.shell_mode = True
        return review_manager
    except (TypeError, KeyError):
        review_manager = colrev.review_manager.ReviewManager(**review_manager_params)
        ctx.obj = {"review_manager": review_manager}
        if SHELL_MODE:
            review_manager.shell_mode = True
        return review_manager


@main.command(help_priority=100)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def shell(
    ctx: click.core.Context,
) -> None:
    """Starts a interactive terminal"""
    from prompt_toolkit.history import FileHistory

    print(f"CoLRev version {colrev.__version__}")
    print("Workflow: status > OPERATION > validate")
    print("https://colrev-environment.github.io/colrev/")
    print("Type exit to close the shell")
    print()
    prompt_kwargs = {"history": FileHistory(".history"), "message": "CoLRev > "}
    # pylint: disable=global-statement
    global SHELL_MODE
    SHELL_MODE = True
    click_repl.repl(ctx, prompt_kwargs=prompt_kwargs)


@main.command(name="exit", help_priority=100)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def exit_command(
    ctx: click.core.Context,
) -> None:
    """Starts a interactive terminal"""
    import inspect

    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    if calframe[7].function == "shell":
        raise click_repl.exceptions.ExitReplException


@main.command(help_priority=1)
@click.option(
    "--type",
    "review_type",
    default="colrev.literature_review",
    help="Review type for the setup.",
    is_flag=False,
    flag_value="add_interactively",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.option(
    "--light",
    is_flag=True,
    default=False,
    help="Setup a lightweight repository (without Docker services)",
)
@click.option(
    "--example",
    is_flag=True,
    default=False,
    help="Add search results example",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def init(
    ctx: click.core.Context,
    review_type: str,
    example: bool,
    force: bool,
    light: bool,
) -> None:
    """Initialize (define review objectives and type)

    Docs: https://colrev-environment.github.io/colrev/manual/problem_formulation/init.html
    """
    import colrev.ops.init

    review_type = _add_endpoint_interactively(review_type, EndpointType.review_type)
    colrev.ops.init.Initializer(
        review_type=review_type,
        target_path=Path.cwd(),
        example=example,
        force_mode=force,
        light=light,
        exact_call=EXACT_CALL,
    )


@main.command(help_priority=2)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def status(
    ctx: click.core.Context,
    verbose: bool,
) -> None:
    """Show status"""
    try:
        review_manager = get_review_manager(
            ctx,
            {"force_mode": False, "verbose_mode": verbose, "exact_call": EXACT_CALL},
        )
        status_operation = review_manager.get_status_operation()
        colrev.ui_cli.cli_status_printer.print_project_status(status_operation)

    except KeyboardInterrupt:
        print("Stopped...")
    except colrev_exceptions.RepoSetupError as exc:
        print(exc)


@main.command(help_priority=100)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.pass_context
def dashboard(
    ctx: click.core.Context,
    verbose: bool,
) -> None:
    """Allows to track project progress through dashboard"""
    import colrev.packages.ui_web.src.dashboard

    try:
        colrev.packages.ui_web.src.dashboard.main()
    except colrev_exceptions.NoRecordsError:
        print("No records imported yet.")
    except colrev_exceptions.CoLRevException as exc:
        if verbose:
            raise exc
        print(exc)


@main.command(help_priority=3)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def retrieve(
    ctx: click.core.Context,
    verbose: bool,
    force: bool,
) -> None:
    """Retrieve, a high-level operation, consists of search, load, prep, and dedupe.

    \b
    To retrieve search results:
    - Copy files (*.bib, *.ris, *.xlsx, ...) to the directory data/search or
    - Copy PDF files to the directory data/pdfs or
    - Add an API-based search, as described in the documentation:

    https://colrev-environment.github.io/colrev/manual/metadata_retrieval/search.html
    """

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force, "high_level_operation": True},
    )

    pdf_dir = review_manager.paths.pdf
    search_dir = review_manager.paths.search
    if not any(search_dir.iterdir()) and not any(pdf_dir.iterdir()):
        # Note : API-based searches automatically retrieve files
        # when they are added, i.e., the following message should
        # not be shown.
        print(
            "To retrieve search results,\n"
            " - copy files (*.bib, *.ris, *.xlsx, ...) "
            f"to the directory {review_manager.paths.SEARCH_DIR} or\n"
            f" - copy PDF files to the directory {review_manager.paths.PDF_DIR} or \n"
            " - add an API-based search, as described in the documentation:\n"
            "https://colrev-environment.github.io/colrev/manual/metadata_retrieval/search.html"
        )
        return

    review_manager.logger.info("Retrieve")
    review_manager.logger.info(
        "Retrieve is a high-level operation consisting of search, load, prep, and dedupe:"
    )
    print()

    review_manager.exact_call = "colrev search"
    search_operation = review_manager.get_search_operation()
    search_operation.add_most_likely_sources()
    search_operation.main(rerun=False)

    print()

    review_manager.exact_call = "colrev load"
    load_operation = review_manager.get_load_operation(hide_load_explanation=True)
    load_operation.main(keep_ids=False)

    print()
    review_manager.exact_call = "colrev prep"
    prep_operation = review_manager.get_prep_operation()
    prep_operation.main()

    print()

    review_manager.exact_call = "colrev dedupe"
    dedupe_operation = review_manager.get_dedupe_operation()
    dedupe_operation.main()


@main.command(help_priority=4)
@click.option(
    "-a",
    "--add",
    help="""Search source to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option("-p", "--params", type=str, help="Parameters", default="")
@click.option("--view", is_flag=True, default=False, help="View search sources")
@click.option(
    "-s",
    "--selected",
    is_flag=False,
    flag_value="select_interactively",
    default=None,
    help="Only retrieve search results for selected sources",
)
@click.option(
    "-r",
    "--rerun",
    is_flag=True,
    default=False,
    help="Rerun API-based searches, retrieving and updating all records "
    + "(not just the most recent ones)",
)
@click.option(
    "-bws",
    help="Backward search on a selected paper",
)
@click.option(
    "--skip",
    is_flag=True,
    default=False,
    help="Skip adding new SearchSources",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom search script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def search(
    ctx: click.core.Context,
    add: str,
    params: str,
    view: bool,
    selected: str,
    rerun: bool,
    bws: str,
    skip: str,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Search for records

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_retrieval/search.html
    """

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )

    search_operation = review_manager.get_search_operation()

    if view:
        for source in search_operation.sources:
            search_operation.review_manager.p_printer.pprint(source)
        return

    if add:
        # pylint: disable=import-outside-toplevel
        # pylint: disable=reimported
        import colrev.ui_cli.add_package_to_settings

        add = _add_endpoint_interactively(add, EndpointType.search_source)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=search_operation,
            package_identifier=add,
            params=params,
        )
        return

    if skip:
        existing_sources = ",".join(
            str(
                source.filename
                for source in search_operation.review_manager.settings.sources
            )
        )
        search_operation.main(selection_str=existing_sources, rerun=rerun)
        return

    if setup_custom_script:
        import colrev.ui_cli.setup_custom_scripts

        colrev.ui_cli.setup_custom_scripts.setup_custom_search_script(
            review_manager=review_manager
        )
        print("Activated custom_search_script.py.")
        print(
            f"Please update the source in {review_manager.paths.SETTINGS_FILE} and commit."
        )
    elif bws:
        import colrev.ui_cli.search_backward_selective

        colrev.ui_cli.search_backward_selective.main(
            search_operation=search_operation, bws=bws
        )
        return

    import colrev.ui_cli.detect_and_add_source

    cli_source_adder = colrev.ui_cli.detect_and_add_source.CLISourceAdder(
        search_operation=search_operation
    )
    sources_added = cli_source_adder.add_new_sources()
    if sources_added:
        if selected is None:
            selected = ",".join(
                [
                    str(source.filename)
                    for source in search_operation.review_manager.settings.sources
                    if source.filename not in [x.filename for x in sources_added]
                ]
            )
        else:
            input(selected)  # notify /handle pre-selected when adding new
    else:
        selected = _select_source_interactively(selected, review_manager)

    search_operation.main(selection_str=selected, rerun=rerun)


@main.command(help_priority=5)
@click.option(
    "-k",
    "--keep_ids",
    is_flag=True,
    default=False,
    help="Do not change the record IDs. Useful when importing an existing sample.",
)
@click.option(
    "-sq",
    "--skip_query",
    is_flag=True,
    default=False,
    help="Skip entering the search query (if applicable)",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def load(
    ctx: click.core.Context,
    keep_ids: bool,
    skip_query: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Load records

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_retrieval/load.html
    """

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )
    load_operation = review_manager.get_load_operation()

    # Note : reinitialize to load new scripts:
    load_operation = review_manager.get_load_operation(hide_load_explanation=True)

    load_operation.main(keep_ids=keep_ids)


@main.command(help_priority=6)
@click.option(
    "-a",
    "--add",
    help="""Prep package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "-k",
    "--keep_ids",
    is_flag=True,
    default=False,
    help="Do not change the record IDs. Useful when importing an existing sample.",
)
@click.option(
    "--polish",
    is_flag=True,
    default=False,
    help="Polish record metadata (includes records in md_processed or beyond).",
)
@click.option(
    "-d",
    "--debug",
    type=str,
    help="Debug the preparation step for a selected record (can be 'all').",
)
@click.option(
    "--cpu",
    type=int,
    help="Number of cpus (parallel processes)",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom prep script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def prep(
    ctx: click.core.Context,
    add: str,
    params: str,
    keep_ids: bool,
    polish: bool,
    debug: str,
    cpu: int,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Prepare records

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_retrieval/prep.html
    """

    try:
        review_manager = get_review_manager(
            ctx,
            {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL},
        )

        if debug:
            review_manager.force_mode = True
            debug_prep_operation = review_manager.get_prep_operation(
                polish=polish,
                cpu=cpu,
                debug=True,
            )
            debug_prep_operation.run_debug(debug_ids=debug)  # type: ignore
            return

        prep_operation = review_manager.get_prep_operation(polish=polish, cpu=cpu)
        if setup_custom_script:
            prep_operation.setup_custom_script()
            print("Activated custom_prep_script.py.")
            print(
                "Please check and adapt its position in the "
                f"{review_manager.paths.SETTINGS_FILE} and commit."
            )
            return
        if add:
            add = _add_endpoint_interactively(add, EndpointType.prep)
            colrev.ui_cli.add_package_to_settings.add_package_to_settings(
                PACKAGE_MANAGER,
                operation=prep_operation,
                package_identifier=add,
                params=params,
            )
            return

        prep_operation.main(keep_ids=keep_ids)

    except colrev_exceptions.ServiceNotAvailableException as exc:
        print(exc)
        print("You can use the force mode to override")
        print(f"  {Colors.ORANGE}colrev prep -f{Colors.END}")
        return


@main.command(help_priority=7)
@click.option(
    "-a",
    "--add",
    help="""Prep-man script  to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "--stats",
    is_flag=True,
    default=False,
    help="Print statistics of records with colrev_status md_needs_manual_preparation",
)
@click.option(
    "-l",
    "--languages",
    is_flag=True,
    default=False,
    help="Export spreadsheet to add missing language fields.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def prep_man(
    ctx: click.core.Context,
    add: str,
    params: str,
    stats: bool,
    languages: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Prepare records manually

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_retrieval/prep.html
    """

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )
    prep_man_operation = review_manager.get_prep_man_operation()
    if languages:
        prep_man_operation.prep_man_langs()
        return

    if stats:
        prep_man_operation.prep_man_stats()
        return

    if add:
        add = _add_endpoint_interactively(add, EndpointType.prep_man)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=prep_man_operation,
            package_identifier=add,
            params=params,
        )
        return

    prep_man_operation.main()


def _view_dedupe_details(dedupe_operation: colrev.ops.dedupe.Dedupe) -> None:
    info = dedupe_operation.get_info()

    if len(info["same_source_merges"]) > 0:
        print(f"\n\n{Colors.RED}Same source merges to check:{Colors.END}")
        print("\n- " + "\n- ".join(info["same_source_merges"]))

    print(info["source_overlaps"])


@main.command(help_priority=8)
@click.option(
    "-a",
    "--add",
    help="""Dedupe package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "-m",
    "--merge",
    help="Merge records by providing a comma-separated list of IDs (ID1,ID2).",
    required=False,
)
@click.option(
    "-u",
    "--unmerge",
    help="Unmerge records by providing a comma-separated list of IDs (ID1,ID2).",
    required=False,
)
@click.option(
    "-f",
    "--fix_errors",
    is_flag=True,
    default=False,
    help="Fix errors marked in duplicates_to_validate.xlsx "
    "or non_duplicates_to_validate.xlsx or "
    "a dupes.txt file containing comma-separated ID tuples",
)
@click.option(
    "-gid",
    help="Merge records with identical global IDs (e.g., DOI)",
    is_flag=True,
    default=False,
)
@click.option("--view", is_flag=True, default=False, help="View dedupe info")
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    default=False,
    help="Debug mode",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def dedupe(
    ctx: click.core.Context,
    add: str,
    params: str,
    merge: str,
    unmerge: str,
    fix_errors: bool,
    gid: bool,
    view: bool,
    debug: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Deduplicate records

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_retrieval/dedupe.html
    """

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )
    state_transition_operation = not view
    dedupe_operation = review_manager.get_dedupe_operation(
        notify_state_transition_operation=state_transition_operation
    )

    if add:
        add = _add_endpoint_interactively(add, EndpointType.dedupe)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=dedupe_operation,
            package_identifier=add,
            params=params,
        )
        return

    if merge:
        assert "," in merge
        merge_ids = [i.split(",") for i in merge.split(";")]

        dedupe_operation.merge_records(merge=merge_ids)
        return

    if unmerge:
        dedupe_operation.unmerge_records(current_record_ids=unmerge.split(","))
        return
    if gid:
        dedupe_operation.merge_based_on_global_ids(apply=True)
        return
    if fix_errors:
        review_manager.report_logger.info("Dedupe: fix errors")
        review_manager.logger.info("Dedupe: fix errors")
        if not (
            dedupe_operation.dupe_file.is_file()
            or dedupe_operation.non_dupe_file_xlsx.is_file()
            or dedupe_operation.non_dupe_file_txt.is_file()
        ):
            review_manager.logger.error("No file with potential errors found.")
            return

        false_positives = colrev.ui_cli.dedupe_errors.load_dedupe_false_positives(
            dedupe_operation=dedupe_operation
        )
        false_negatives = colrev.ui_cli.dedupe_errors.load_dedupe_false_negatives(
            dedupe_operation=dedupe_operation
        )

        dedupe_operation.fix_errors(
            false_positives=false_positives,
            false_negatives=false_negatives,
        )
        print(
            "You can manually remove the duplicates_to_validate.xlsx, "
            "non_duplicates_to_validate.xlsx, and dupes.txt files."
        )
        return

    if view:
        _view_dedupe_details(dedupe_operation)
        return

    dedupe_operation.main(debug=debug)


@main.command(help_priority=9)
@click.option(
    "-a",
    "--add",
    help="""Prescreen package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "--include_all",
    is_flag=True,
    default=False,
    help="Include all records in prescreen",
)
@click.option(
    "--include_all_always",
    is_flag=True,
    default=False,
    help="Include all records in this and future prescreens",
)
@click.option(
    "--export_format",
    type=click.Choice(["CSV", "XLSX"], case_sensitive=False),
    help="Export table with the screening decisions",
)
@click.option(
    "--import_table",
    type=click.Path(exists=True),
    help="Import file with the screening decisions (csv/xlsx supported)",
)
@click.option(
    "--create_split",
    type=int,
    help="Split the prescreen between n researchers "
    + "(same size, non-overlapping samples)",
)
@click.option(
    "--split",
    type=str,
    default="",
    help="Prescreen a split sample",
)
@click.option(
    "-i",
    "--include",
    help="Prescreen include records based on IDs (ID1,ID2,...).",
    required=False,
)
@click.option(
    "-e",
    "--exclude",
    help="Prescreen exclude records based on IDs (ID1,ID2,...).",
    required=False,
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom prescreen script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def prescreen(
    ctx: click.core.Context,
    add: str,
    params: str,
    include_all: bool,
    include_all_always: bool,
    export_format: str,
    import_table: str,
    create_split: int,
    split: str,
    include: str,
    exclude: str,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Pre-screen exclusion based on metadata (titles and abstracts)

    Docs: https://colrev-environment.github.io/colrev/manual/metadata_prescreen/prescreen.html
    """

    # pylint: disable=too-many-locals

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )
    prescreen_operation = review_manager.get_prescreen_operation()

    if export_format:
        prescreen_operation.export_table(export_table_format=export_format)

    elif import_table:
        prescreen_operation.import_table(import_table_path=import_table)

    elif include_all or include_all_always:
        prescreen_operation.include_all_in_prescreen(persist=include_all_always)

    elif create_split:
        splits = prescreen_operation.create_prescreen_split(create_split=create_split)
        for created_split in splits:
            print(created_split + "\n")

    elif include:
        prescreen_operation.include_records(ids=include)

    elif exclude:
        prescreen_operation.exclude_records(ids=exclude)

    elif setup_custom_script:
        prescreen_operation.setup_custom_script()
        print("Activated custom_prescreen_script.py.")
    elif add:
        add = _add_endpoint_interactively(add, EndpointType.prescreen)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=prescreen_operation,
            package_identifier=add,
            params=params,
        )
    else:
        review_manager.logger.info("Prescreen")
        review_manager.logger.info(
            "Exclude irrelevant records based on metadata (i.e., titles and abstracts)."
        )
        review_manager.logger.info("Remaining records are retained provisionally")
        review_manager.logger.info(
            "In the screen, they can be included or excluded based on full-text documents."
        )
        review_manager.logger.info(
            "See https://colrev-environment.github.io/"
            "colrev/manual/metadata_prescreen/prescreen.html"
        )

        prescreen_operation.main(split_str=split)


@main.command(help_priority=10)
@click.option(
    "-a",
    "--add",
    help="""Screen package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "--include_all",
    is_flag=True,
    default=False,
    help="Include all records in the screen",
)
@click.option(
    "--include_all_always",
    is_flag=True,
    default=False,
    help="Include all records in this and future screens",
)
@click.option(
    "-ac",
    "--add_criterion",
    help="Add a screening criterion. "
    "Format: -ac 'criterion_name,criterion explanation'",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-dc",
    "--delete_criterion",
    type=str,
    help="Delete a screening criterion. Format: -dc 'criterion_name'",
)
@click.option(
    "--create_split",
    type=int,
    help="Split the screen between n researchers "
    + "(each researcher screens the same number of papers without overlaps)",
)
@click.option(
    "--split",
    type=str,
    default="",
    help="Screen a split sample",
)
@click.option(
    "-abstracts",
    "--add_abstracts_from_tei",
    is_flag=True,
    default=False,
    help="Add abstracts from TEI files",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom screen script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def screen(
    ctx: click.core.Context,
    add: str,
    params: str,
    include_all: bool,
    include_all_always: bool,
    add_criterion: str,
    delete_criterion: str,
    create_split: int,
    split: str,
    add_abstracts_from_tei: bool,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Screen based on PDFs and inclusion/exclusion criteria

    Docs: https://colrev-environment.github.io/colrev/manual/pdf_screen/screen.html
    """

    review_manager = get_review_manager(
        ctx, {"verbose_mode": verbose, "force_mode": force, "exact_call": EXACT_CALL}
    )
    screen_operation = review_manager.get_screen_operation()

    if add_abstracts_from_tei:
        screen_operation.add_abstracts_from_tei()

    if add:
        add = _add_endpoint_interactively(add, EndpointType.screen)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=screen_operation,
            package_identifier=add,
            params=params,
        )
        return

    if include_all or include_all_always:
        screen_operation.include_all_in_screen(persist=include_all_always)
        return
    if add_criterion:
        _add_screen_criterion_interactively(add_criterion, screen_operation)

    if delete_criterion:
        screen_operation.delete_criterion(delete_criterion)
        return
    if create_split:
        splits = screen_operation.create_screen_split(create_split=create_split)
        for created_split in splits:
            print(created_split + "\n")
        return
    if setup_custom_script:
        screen_operation.setup_custom_script()
        print("Activated custom_screen_script.py.")
        return

    screen_operation.main(split_str=split)


def _extract_coverpage(*, cover: Path) -> None:
    # pylint: disable=import-outside-toplevel
    import colrev.record.record_pdf

    Filepaths.COVERPAGES.mkdir(exist_ok=True)

    assert Path(cover).suffix == ".pdf"
    record = colrev.record.record_pdf.PDFRecord(
        {Fields.FILE: cover}, path=Path(cover).parent
    )
    record.extract_pages(pages=[0], save_to_path=Filepaths.COVERPAGES)


@main.command(help_priority=17)
@click.argument("path", nargs=1, type=click.Path(exists=True))
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdf(
    ctx: click.core.Context,
    path: str,
) -> None:
    """Process a PDF"""

    ret = ""
    while ret in ["c", "h", ""]:
        ret = input("Option (c: remove cover page, h: show hashes, q: quit)")
        if ret == "c":
            _extract_coverpage(cover=Path(path))
        elif ret == "h":
            _print_pdf_hashes(pdf_path=Path(path))
        # elif ret == "o":
        #     print("TODO : ocr")
        # elif ret == "r":
        #     print("TODO: remove comments")
        # elif ret == "m":
        #     print("TODO : extract metadata")
        # elif ret == "t":
        #     print("TODO : create tei")
        # elif ret == "i":
        #     print("TODO: print infos (website / retracted /...)")


@main.command(help_priority=11)
@click.option(
    "--discard",
    is_flag=True,
    default=False,
    help="Discard all missing PDFs as not_available",
)
@click.option(
    "-d",
    "--dir",
    "open_dir",
    is_flag=True,
    default=False,
    help="Open the PDFs directory",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdfs(
    ctx: click.core.Context,
    discard: bool,
    open_dir: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Retrieve and prepare PDFs"""

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
            "high_level_operation": True,
        },
    )

    if open_dir:
        # pylint: disable=consider-using-with
        # pylint: disable=no-member

        path = review_manager.path / Path("data/pdfs")
        webbrowser.open(str(path))
        return

    if discard:
        pdf_prep_man_operation = review_manager.get_pdf_prep_man_operation()
        pdf_prep_man_operation.discard()

        pdf_get_man_operation = review_manager.get_pdf_get_man_operation()
        pdf_get_man_operation.discard()

        return

    review_manager.logger.info("PDFs")
    review_manager.logger.info(
        "PDFs is a high-level operation consisting of pdf-get and pdf-prep:"
    )
    print()

    pdf_get_operation = review_manager.get_pdf_get_operation(
        notify_state_transition_operation=True
    )
    pdf_get_operation.main()


@main.command(help_priority=12)
@click.option(
    "-a",
    "--add",
    help="""PDF-get package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "-c",
    "--copy-to-repo",
    is_flag=True,
    default=False,
    help="Copy PDF files to the repository (the /pdfs directory)",
)
@click.option(
    "-r",
    "--rename",
    is_flag=True,
    default=False,
    help="Rename the PDF files according to record IDs",
)
@click.option(
    "--relink_pdfs",
    is_flag=True,
    default=False,
    help="Recreate links to PDFs based on colrev pdf-IDs (when PDFs were renamed)",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom pdf-get script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdf_get(
    ctx: click.core.Context,
    add: str,
    params: str,
    copy_to_repo: bool,
    rename: bool,
    relink_pdfs: bool,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Get PDFs

    Docs: https://colrev-environment.github.io/colrev/manual/pdf_retrieval/pdf_get.html
    """

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )

    state_transition_operation = not relink_pdfs and not setup_custom_script
    pdf_get_operation = review_manager.get_pdf_get_operation(
        notify_state_transition_operation=state_transition_operation
    )

    if add:
        add = _add_endpoint_interactively(add, EndpointType.pdf_get)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=pdf_get_operation,
            package_identifier=add,
            params=params,
        )
        return
    if relink_pdfs:
        pdf_get_operation.relink_pdfs()
        return
    if copy_to_repo:
        pdf_get_operation.copy_pdfs_to_repo()
        return
    if rename:
        pdf_get_operation.rename_pdfs()
        return
    if setup_custom_script:
        pdf_get_operation.setup_custom_script()
        print("Activated custom_pdf_get_script.py.")
        return

    pdf_get_operation.main()


@main.command(help_priority=13)
@click.option(
    "-a",
    "--add",
    help="""PDF-get-man package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "-e",
    "--export",
    is_flag=True,
    default=False,
    help="Export a table.",
)
@click.option(
    "--discard",
    is_flag=True,
    default=False,
    help="Discard all missing PDFs as not_available",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdf_get_man(
    ctx: click.core.Context,
    add: str,
    params: str,
    export: bool,
    discard: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Get PDFs manually

    Docs: https://colrev-environment.github.io/colrev/manual/pdf_retrieval/pdf_get.html
    """

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )
    pdf_get_man_operation = review_manager.get_pdf_get_man_operation()

    if add:
        add = _add_endpoint_interactively(add, EndpointType.pdf_get_man)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=pdf_get_man_operation,
            package_identifier=add,
            params=params,
        )
        return

    if export:
        records = pdf_get_man_operation.review_manager.dataset.load_records_dict()
        pdf_get_man_records = [
            r
            for r in records.values()
            if r[Fields.STATUS]
            in [
                RecordState.pdf_needs_manual_retrieval,
                RecordState.rev_prescreen_included,
            ]
        ]
        pdf_get_man_records_df = pd.DataFrame.from_records(pdf_get_man_records)
        pdf_get_man_records_df = pdf_get_man_records_df[
            pdf_get_man_records_df.columns.intersection(
                [
                    Fields.ID,
                    Fields.AUTHOR,
                    Fields.YEAR,
                    Fields.TITLE,
                    Fields.JOURNAL,
                    Fields.BOOKTITLE,
                    Fields.VOLUME,
                    Fields.NUMBER,
                    Fields.URL,
                    Fields.DOI,
                ]
            )
        ]
        pdf_get_man_records_df.to_csv("pdf_get_man_records.csv", index=False)
        pdf_get_man_operation.review_manager.logger.info(
            "Created pdf_get_man_records.csv"
        )
        return
    if discard:
        pdf_get_man_operation.discard()
        return

    pdf_get_man_operation.main()


def _print_pdf_hashes(*, pdf_path: Path) -> None:

    import colrev.record.record_pdf
    import pymupdf

    doc = pymupdf.Document(pdf_path)
    last_page_nr = doc.page_count

    assert Path(pdf_path).suffix == ".pdf"
    record = colrev.record.record_pdf.PDFRecord(
        {Fields.FILE: pdf_path.name, Fields.ID: "NA"}, path=pdf_path.parents[0]
    )
    first_page_average_hash_16 = record.get_pdf_hash(page_nr=1, hash_size=16)
    print(f"first page: {first_page_average_hash_16}")
    first_page_average_hash_32 = record.get_pdf_hash(page_nr=1, hash_size=32)
    print(f"first page: {first_page_average_hash_32}")

    last_page_average_hash_16 = record.get_pdf_hash(page_nr=last_page_nr, hash_size=16)
    print(f"last page: {last_page_average_hash_16}")
    last_page_average_hash_32 = record.get_pdf_hash(page_nr=last_page_nr, hash_size=32)
    print(f"last page: {last_page_average_hash_32}")


@main.command(help_priority=14)
@click.option(
    "-a",
    "--add",
    help="""PDF-prep package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "--update_colrev_pdf_ids", is_flag=True, default=False, help="Update colrev_pdf_ids"
)
@click.option(
    "-b",
    "--batch_size",
    required=False,
    type=int,
    default=0,
    help="Batch size (when not all records should be processed in one batch).",
)
@click.option(
    "--reprocess",
    is_flag=True,
    default=False,
    help="Prepare all PDFs again (pdf_needs_manual_preparation).",
)
@click.option(
    "--tei",
    is_flag=True,
    default=False,
    help="Generate TEI documents.",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom pdf-prep script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdf_prep(
    ctx: click.core.Context,
    add: str,
    params: str,
    batch_size: int,
    update_colrev_pdf_ids: bool,
    reprocess: bool,
    setup_custom_script: bool,
    tei: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Prepare PDFs

    Docs: https://colrev-environment.github.io/colrev/manual/pdf_retrieval/pdf_prep.html
    """

    # pylint: disable=import-outside-toplevel

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )
    pdf_prep_operation = review_manager.get_pdf_prep_operation(reprocess=reprocess)

    if add:
        add = _add_endpoint_interactively(add, EndpointType.pdf_prep)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=pdf_prep_operation,
            package_identifier=add,
            params=params,
        )
        return

    try:
        if update_colrev_pdf_ids:
            pdf_prep_operation.update_colrev_pdf_ids()

        elif setup_custom_script:
            pdf_prep_operation.setup_custom_script()
            print("Activated custom_pdf_prep_script.py.")
        elif tei:
            pdf_prep_operation.generate_tei()
        else:
            pdf_prep_operation.main(batch_size=batch_size)

    except KeyboardInterrupt:
        print("Stopped the process")


def _delete_first_pages_cli(
    pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan, record_id: str
) -> None:
    records = pdf_prep_man_operation.review_manager.dataset.load_records_dict()
    while True:
        if record_id in records:
            record_dict = records[record_id]
            if Fields.FILE in record_dict:
                print(record_dict[Fields.FILE])
                pdf_path = pdf_prep_man_operation.review_manager.path / Path(
                    record_dict[Fields.FILE]
                )
                pdf_prep_man_operation.extract_coverpage(filepath=pdf_path)
                pdf_prep_man_operation.set_pdf_man_prepared(
                    colrev.record.record.Record(record_dict)
                )
            else:
                print("no file in record")
        if input("Extract coverpage from another PDF? (y/n)") == "n":
            break
        record_id = input("ID of next PDF for coverpage extraction:")


@main.command(help_priority=15)
@click.option(
    "-a",
    "--add",
    help="""PDF-prep-man package to be added.""",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "-dfp",
    "--delete_first_page",
    type=str,
    help="Delete first page of PDF. Format: --delete_first_page ID",
)
@click.option(
    "--stats",
    is_flag=True,
    default=False,
    help="Print statistics of records with colrev_status pdf_needs_manual_preparation",
)
@click.option(
    "--discard",
    is_flag=True,
    default=False,
    help="Discard records whose PDF needs to be prepared manually",
)
@click.option(
    "--extract",
    is_flag=True,
    default=False,
    help="Extract records for manual_preparation (to csv and bib)",
)
@click.option(
    "--apply",
    is_flag=True,
    default=False,
    help="Apply manual preparation (from csv or bib)",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pdf_prep_man(
    ctx: click.core.Context,
    add: str,
    params: str,
    delete_first_page: str,
    stats: bool,
    discard: bool,
    extract: bool,
    apply: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Prepare PDFs manually

    Docs: https://colrev-environment.github.io/colrev/manual/pdf_retrieval/pdf_prep.html
    """

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )
    pdf_prep_man_operation = review_manager.get_pdf_prep_man_operation()

    if add:
        add = _add_endpoint_interactively(add, EndpointType.pdf_prep_man)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=pdf_prep_man_operation,
            package_identifier=add,
            params=params,
        )
    if delete_first_page:
        _delete_first_pages_cli(pdf_prep_man_operation, delete_first_page)
        return
    if discard:
        pdf_prep_man_operation.discard()
        return
    if stats:
        pdf_prep_man_operation.pdf_prep_man_stats()
        return
    if extract:
        pdf_prep_man_operation.extract_needs_pdf_prep_man()
        return
    if apply:
        pdf_prep_man_operation.apply_pdf_prep_man()
        return

    pdf_prep_man_operation.main()


@main.command(help_priority=16)
@click.option(
    "-a",
    "--add",
    help="Data package to be added.",
    is_flag=False,
    flag_value="add_interactively",
    default="",
)
@click.option(
    "-p",
    "--params",
    type=str,
    help="Parameters",
)
@click.option(
    "--reading_heuristics",
    is_flag=True,
    default=False,
    help="Heuristics to prioritize reading efforts",
)
@click.option(
    "-scs",
    "--setup_custom_script",
    is_flag=True,
    default=False,
    help="Setup template for custom data script.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def data(
    ctx: click.core.Context,
    add: str,
    params: str,
    reading_heuristics: bool,
    setup_custom_script: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Complete selected forms of data analysis and synthesis

    Docs: https://colrev-environment.github.io/colrev/manual/data/data.html
    """

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )
    data_operation = review_manager.get_data_operation()

    if reading_heuristics:
        heuristic_results = data_operation.reading_heuristics()
        review_manager.p_printer.pprint(heuristic_results)
        return
    if setup_custom_script:
        data_operation.setup_custom_script()
        print("Activated custom_data_script.py.")
        print(
            f"Please update the data_format in {review_manager.paths.SETTINGS_FILE} and commit."
        )
        return

    if add:
        add = _add_endpoint_interactively(add, EndpointType.data)
        colrev.ui_cli.add_package_to_settings.add_package_to_settings(
            PACKAGE_MANAGER,
            operation=data_operation,
            package_identifier=add,
            params=params,
        )
        return

    ret = data_operation.main()
    if data_operation.review_manager.in_ci_environment():
        if ret["ask_to_commit"]:
            review_manager.dataset.create_commit(
                msg="Data and synthesis", manual_author=True
            )
    else:
        if ret["ask_to_commit"]:
            if input("Create commit (y/n)?") == "y":
                review_manager.dataset.create_commit(
                    msg="Data and synthesis", manual_author=True
                )
        if ret["no_endpoints_registered"]:
            print(
                "No data format not specified. "
                "To register a data endpoint, "
                "use one (or several) of the following \n"
                "    colrev data --add colrev.paper_md\n"
                "    colrev data --add colrev.structured\n"
                "    colrev data --add colrev.bibliography_export\n"
                "    colrev data --add colrev.prisma\n"
                "    colrev data --add colrev.github_pages\n"
                "    colrev data --add colrev.zettlr\n"
                "    colrev data --add colrev.colrev_curation"
            )


@main.command(help_priority=17)
@click.argument("scope", nargs=1)
@click.option(
    "--filter",
    "filter_settings",
    type=click.Choice(["prepare", "dedupe", "merge", "all"], case_sensitive=False),
    default="all",
    help="prepare, merge, or all.",
)
@click.option(
    "--threshold",
    type=float,
    default=0.05,
    help="Change score threshold for changes to display.",
)
@click.option(
    "--properties",
    is_flag=True,
    default=False,
    help="Flag indicating whether to validate the review properties.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def validate(
    ctx: click.core.Context,
    scope: str,
    filter_settings: str,
    threshold: float,
    properties: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Validate changes in the given commit (scope)

    \b
    The validation scope argument can be
    - A commit-sha,
    - '.' for the latest commit,
    - HEAD~4 for commit 4 before HEAD
    - A contributor name
    """

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
            "exact_call": EXACT_CALL,
        },
    )
    validate_operation = review_manager.get_validate_operation()

    validation_details = validate_operation.main(
        scope=scope,
        filter_setting=filter_settings,
        properties=properties,
    )

    if validation_details:
        colrev.ui_cli.cli_validation.validate(
            validate_operation=validate_operation,
            validation_details=validation_details,
            threshold=threshold,
        )

    review_manager.logger.info("%sCompleted validation%s", Colors.GREEN, Colors.END)


@main.command(help_priority=18)
@click.option(
    "--id",  # pylint: disable=invalid-name
    "record_id",
    help="Record ID to trace (citation_key).",
    required=True,
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def trace(
    ctx: click.core.Context,
    record_id: str,  # pylint: disable=invalid-name
) -> None:
    """Trace a record"""

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": False,
            "force_mode": False,
            "exact_call": EXACT_CALL,
        },
    )
    trace_operation = review_manager.get_trace_operation()
    trace_operation.main(record_id=record_id)


def _select_target_repository(environment_registry: list) -> Path:
    while True:
        for i, local_source in enumerate(environment_registry):
            print(
                f"{i+1} - {local_source['repo_name']} ({local_source['repo_source_path']})"
            )
        sel_str = input("Select target repository: ")
        sel = int(sel_str) - 1
        if sel in range(0, len(environment_registry)):
            target = Path(environment_registry[sel]["repo_source_path"])
            return target


@main.command(help_priority=19)
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True),
    help="Path to file(s)",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def distribute(ctx: click.core.Context, path: Path, verbose: bool, force: bool) -> None:
    """Distribute records to other local repositories"""

    if not path:
        path = Path.cwd()
    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": force,
        },
    )
    distribute_operation = review_manager.get_distribute_operation()
    environment_registry = distribute_operation.get_environment_registry()

    target = _select_target_repository(environment_registry=environment_registry)
    # Note : add a "distribution mode" option?
    # (whole file -> add as source/load vs. records individually like a prescreen)
    distribute_operation.main(path=path, target=target)


def _print_environment_status(
    review_manager: colrev.review_manager.ReviewManager,
) -> None:
    environment_manager = review_manager.get_environment_manager()
    environment_details = environment_manager.get_environment_details()

    print("\nCoLRev environment status\n")
    print("Index\n")
    if environment_details["index"]["status"] == "up":
        print(f" - Status: {Colors.GREEN}up{Colors.END}")
        print(f' - Path          : {environment_details["index"]["path"]}')
        print(f' - Size          : {environment_details["index"]["size"]} records')
        print(f' - Last modified : {environment_details["index"]["last_modified"]}')
    else:
        print(f" - Status: {Colors.RED}down{Colors.END}")

    print("\nCoLRev projects\n")
    project_repos = [
        x
        for x in environment_details["local_repos"]["repos"]
        if "curated_metadata" not in x["repo_source_path"]
    ]
    for colrev_repo in sorted(project_repos, key=lambda d: d["repo_name"]):
        repo_stats = f' {colrev_repo["repo_name"]}'
        if colrev_repo["remote"]:
            if colrev_repo["behind_remote"]:
                repo_stats += " (shared, behind remote)"
            else:
                repo_stats += " (shared)"
        print(repo_stats)

        if -1 != colrev_repo["progress"]:
            print(f'    - Progress : {colrev_repo["progress"]*100} %')
        else:
            print("    - Progress : ??")
        print(f'    - Size     : {colrev_repo["size"]} records')
        print(f'    - Path     : {colrev_repo["repo_source_path"]}')

    print("\nCurated CoLRev resources\n")
    curated_repos = [
        x
        for x in environment_details["local_repos"]["repos"]
        if "curated_metadata" in x["repo_source_path"]
    ]
    for colrev_repo in sorted(curated_repos, key=lambda d: d["repo_name"]):
        repo_stats = (
            f' - {colrev_repo["repo_name"].ljust(60, " ")}: '
            f'{str(colrev_repo["size"]).rjust(10, " ")} records'
        )
        if colrev_repo["behind_remote"]:
            repo_stats += " (behind remote)"
        print(repo_stats)

    print("\n")
    if len(environment_details["local_repos"]["broken_links"]) > 0:
        print("Broken links: \n")
        for broken_link in environment_details["local_repos"]["broken_links"]:
            print(f'- {broken_link["repo_source_path"]}')


@main.command(help_priority=20)
@click.option(
    "-i", "--index", is_flag=True, default=False, help="Create the LocalIndex"
)
@click.option(
    "--install",
    help="Install a new resource providing its url "
    + "(e.g., a curated metadata repository)",
)
@click.option("--pull", is_flag=True, default=False, help="Pull curated metadata")
@click.option(
    "-s", "--status", is_flag=True, default=False, help="Print environment status"
)
@click.option(
    "-r",
    "--register",
    is_flag=True,
    default=False,
    help="Register a repository in the CoLRev environment",
)
@click.option(
    "-ur",
    "--unregister",
    type=click.Path(exists=True),
    help="Path of repository to remove from local registry.",
)
@click.option(
    "--update_package_list",
    is_flag=True,
    default=False,
    help="Update the package list (packages).",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.pass_context
def env(
    ctx: click.core.Context,
    index: bool,
    install: str,
    pull: bool,
    status: bool,
    register: bool,
    unregister: bool,
    update_package_list: bool,
    verbose: bool,
) -> None:
    """Manage the environment"""

    # pylint: disable=too-many-branches

    if update_package_list:
        import colrev.package_manager.doc_registry_manager

        doc_reg_manager = (
            colrev.package_manager.doc_registry_manager.DocRegistryManager()
        )
        doc_reg_manager.update()
        return

    if index:
        print("Index rankings")
        import colrev.env.local_index_builder

        local_index_builder = colrev.env.local_index_builder.LocalIndexBuilder(
            verbose_mode=verbose
        )
        local_index_builder.index()
        local_index_builder.index_journal_rankings()
        return

    # The following options may need a review_manager

    review_manager = get_review_manager(
        ctx,
        {
            "verbose_mode": verbose,
            "force_mode": True,
        },
    )

    if install:
        env_resources = review_manager.get_resources()
        if env_resources.install_curated_resource(curated_resource=install):
            print("Successfully installed curated resource.")
            print("To make it available to other projects, run")
            print("colrev env --index")
        return

    if pull:
        environment_manager = review_manager.get_environment_manager()
        for curated_resource in environment_manager.local_repos():
            try:
                curated_resource_path = curated_resource["repo_source_path"]
                if "/curated_metadata/" not in curated_resource_path:
                    continue

                review_manager = get_review_manager(
                    ctx,
                    {
                        "verbose_mode": verbose,
                        "force_mode": False,
                        "path_str": curated_resource_path,
                    },
                )

                review_manager.dataset.pull_if_repo_clean()
                print(f"Pulled {curated_resource_path}")
            except GitCommandError as exc:
                print(exc)
        return

    if status:
        _print_environment_status(review_manager)
        return

    if register:
        environment_manager = review_manager.get_environment_manager()
        environment_manager.register_repo(Path.cwd())
        return

    if unregister is not None:
        environment_manager = review_manager.get_environment_manager()

        local_repos = environment_manager.local_repos()
        if str(unregister) not in [x["source_url"] for x in local_repos]:
            print("Not in local registry (cannot remove): %s", unregister)
        else:
            updated_local_repos = [
                x for x in local_repos if x["source_url"] != str(unregister)
            ]
            environment_manager.environment_registry["local_index"][
                "repos"
            ] = updated_local_repos
            environment_manager.save_environment_registry(
                environment_manager.environment_registry
            )
            logging.info("Removed from local registry: %s", unregister)
        return


@main.command(help_priority=21)
@click.option(
    "--init",
    is_flag=True,
    help="Initialize a CoLRev package",
)
@click.option(
    "--check",
    is_flag=True,
    help="Check a CoLRev package",
)
@click.pass_context
def package(
    ctx: click.core.Context,
    init: bool,
    check: bool,
) -> None:
    """Manage CoLRev packages"""

    if init:
        import colrev.package_manager.init

        colrev.package_manager.init.main()
    if check:
        import colrev.package_manager.check

        colrev.package_manager.check.main()


@main.command(help_priority=22)
# @click.option("-v", "--view", is_flag=True, default=False)
@click.option(
    "-uh",
    "--update_hooks",
    is_flag=True,
    default=False,
    help="Update the pre-commit hooks",
)
@click.option(
    "-m",
    "--modify",
    type=str,
    default="",
    help="Modify the settings through the command line",
)
@click.option(
    "-g",
    "--update-global",
    help="Global settings to update",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def settings(
    ctx: click.core.Context,
    update_hooks: bool,
    modify: str,
    update_global: str,
    verbose: bool,
    force: bool,
) -> None:
    """Settings of the CoLRev project"""

    # pylint: disable=reimported

    from subprocess import check_call  # nosec
    from subprocess import DEVNULL  # nosec
    from subprocess import STDOUT  # nosec
    import json
    import ast
    import glom

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )
    if update_hooks:
        print("Update pre-commit hooks")

        if review_manager.dataset.has_record_changes():
            print("Clean repo required. Commit or stash changes.")
            return

        scripts_to_call = [
            [
                "pre-commit",
                "autoupdate",
                "--repo",
                "https://github.com/CoLRev-Environment/colrev-hooks",
            ],
        ]
        for script_to_call in scripts_to_call:
            check_call(script_to_call, stdout=DEVNULL, stderr=STDOUT)  # nosec

        review_manager.dataset.add_changes(review_manager.paths.PRE_COMMIT_CONFIG)
        review_manager.dataset.create_commit(msg="Update pre-commit hooks")
        print("Successfully updated pre-commit hooks")
        return

    if update_global:
        from colrev.env.environment_manager import EnvironmentManager

        env_man = EnvironmentManager()
        path, value_string = update_global.split("=")
        print(f"Updating registry settings:\n{path} = {value_string}")
        env_man.update_registry(path, value_string)
    if modify:
        # TBD: maybe use glom.delete?
        # There is no simply append...
        # (we could replace the (last) position element with
        # keywords like prescreen.sripts.LAST_POSITION)
        # maybe prescreen.scripts.1.REPLACE/ADD/DELETE = ....
        # modify = 'dedupe.dedupe_package_endpoints='
        # '[{"endpoint":"colrev.simple_dedupe"}]'

        path, value_string = modify.split("=")
        value = ast.literal_eval(value_string)
        review_manager.logger.info("Change settings.%s to %s", path, value)

        with open(review_manager.paths.settings, encoding="utf-8") as file:
            project_settings = json.load(file)

        glom.assign(project_settings, path, value)

        with open(review_manager.paths.settings, "w", encoding="utf-8") as outfile:
            json.dump(project_settings, outfile, indent=4)

        review_manager.dataset.add_changes(review_manager.paths.SETTINGS_FILE)
        review_manager.dataset.create_commit(msg="Change settings", manual_author=True)

    # import colrev_ui.ui_web.settings_editor

    # review_manager = colrev.review_manager.ReviewManager(
    #     force_mode=True, verbose_mode=verbose
    # )
    # settings_operation = colrev.packages.ui_web.src.settings_editor.SettingsEditor(
    #     review_manager=review_manager
    # )
    # settings_operation.open_settings_editor()


@main.command(help_priority=23)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def pull(
    ctx: click.core.Context,
    verbose: bool,
    force: bool,
) -> None:
    """Pull CoLRev project remote and record updates"""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )
    pull_operation = review_manager.get_pull_operation()

    pull_operation.main()


@main.command(help_priority=24)
@click.argument("git_url")
@click.pass_context
def clone(
    ctx: click.core.Context,
    git_url: str,
) -> None:
    """Create local clone from shared CoLRev repository with git_url"""

    import colrev.ops.clone

    clone_operation = colrev.ops.clone.Clone(git_url=git_url)
    clone_operation.clone_git_repo()


@main.command(help_priority=25)
@click.option(
    "-r",
    "--records_only",
    is_flag=True,
    default=False,
    help="Update records only",
)
@click.option(
    "-p",
    "--project_only",
    is_flag=True,
    default=False,
    help="Push project only",
)
@click.option(
    "-a",
    "--all",
    "all_records",
    is_flag=True,
    default=False,
    help="Push record changes/corrections to all sources (not just curations).",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
@catch_exception(handle=(colrev_exceptions.CoLRevException))
def push(
    ctx: click.core.Context,
    records_only: bool,
    project_only: bool,
    all_records: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Push CoLRev project remote and record updates"""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )
    push_operation = review_manager.get_push_operation()

    push_operation.main(
        records_only=records_only, project_only=project_only, all_records=all
    )


def _validate_show(ctx: click.core.Context, param: str, value: str) -> None:
    if value not in ["sample", "settings", "venv"]:
        raise click.BadParameter("Invalid argument")


@main.command(help_priority=26)
@click.argument("keyword")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def show(  # type: ignore
    ctx: click.core.Context,
    keyword: str,
    verbose: bool,
    force: bool,
    callback=_validate_show,
) -> None:
    """Show aspects (sample, ...)"""

    import colrev.process.operation
    import colrev.ui_cli.show_printer

    if keyword == "venv":
        colrev.ui_cli.show_printer.print_venv_notes()
        return

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )

    if keyword == "sample":
        colrev.ui_cli.show_printer.print_sample(review_manager=review_manager)

    elif keyword == "settings":
        print(f"Settings:\n{review_manager.settings}")

    elif keyword == "cmd_history":
        cmds = []
        colrev.ops.check.CheckOperation(review_manager)
        revlist = review_manager.dataset.get_repo().iter_commits()

        for commit in reversed(list(revlist)):
            try:
                cmsg = str(commit.message)
                formatted_date = time.strftime(
                    "%Y-%m-%d %H:%M",
                    time.gmtime(commit.committed_date),
                )
                if not all(x in cmsg for x in ["Command", "Status"]):
                    # pylint: disable=colrev-missed-constant-usage
                    cmsg = "UNKNOWN"
                # min(cmsg.find("Status"), cmsg.find("On commit"))
                if "On commit" in cmsg:
                    cmsg = cmsg[: cmsg.find("On commit")]
                if "Status" in cmsg:
                    cmsg = cmsg[: cmsg.find("Status")]
                commit_message_first_line = (
                    cmsg[cmsg.find("Command") + 8 :]
                    .lstrip()
                    .rstrip()
                    .replace("\n", " ")
                )
                if len(commit_message_first_line) > 800:
                    # pylint: disable=colrev-missed-constant-usage
                    cmsg = "UNKNOWN"
                cmds.append(
                    {
                        "date": formatted_date,
                        "committer": commit.committer.name,
                        "commit_id": commit.hexsha,
                        "cmd": commit_message_first_line,
                    }
                )
            except KeyError:
                continue
        for cmd in cmds:
            print(
                f"{cmd['date']} ({cmd['committer']}, {cmd['commit_id']}):    "
                f"{Colors.ORANGE}{cmd['cmd']}{Colors.END}"
            )


# @main.command(help_priority=28)
# @click.option(
#     "-v",
#     "--verbose",
#     is_flag=True,
#     default=False,
#     help="Verbose: printing more infos",
# )
# @click.option(
#     "-f",
#     "--force",
#     is_flag=True,
#     default=False,
#     help="Force mode",
# )
# @click.pass_context
# def web(
#     ctx: click.core.Context,
#     verbose: bool,
#     force: bool,
# ) -> None:
#     """CoLRev web interface."""

#
#     import colrev.packages.ui_web.src.settings_editor

#     review_manager = colrev.review_manager.ReviewManager(
#         force_mode=force, verbose_mode=verbose
#     )
#     se_instance = colrev.packages.ui_web.src.settings_editor.SettingsEditor(
#         review_manager=review_manager
#     )
#     se_instance.open_settings_editor()


@main.command(hidden=True, help_priority=27)
@click.option(
    "--disable_auto",
    is_flag=True,
    default=False,
    help="Disable automated upgrades",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def upgrade(
    ctx: click.core.Context,
    disable_auto: bool,
    verbose: bool,
    force: bool,
) -> None:
    """Upgrade to the latest CoLRev project version."""

    if disable_auto:
        review_manager = colrev.review_manager.ReviewManager(
            force_mode=True, verbose_mode=verbose, skip_upgrade=True
        )

        review_manager.settings.project.auto_upgrade = False
        review_manager.save_settings()
        review_manager.dataset.create_commit(msg="Disable auto-upgrade")
        return
    review_manager = colrev.review_manager.ReviewManager(
        force_mode=True, verbose_mode=verbose
    )
    upgrade_operation = review_manager.get_upgrade()
    upgrade_operation.main()


@main.command(hidden=True, help_priority=28)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def repare(
    ctx: click.core.Context,
    verbose: bool,
    force: bool,
) -> None:
    """Repare file formatting errors in the CoLRev project."""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )
    repare_operation = review_manager.get_repare()
    repare_operation.main()


@main.command(help_priority=29)
@click.option(
    "--ids",
    help="Remove records and their origins from the repository (ID1,ID2,...).",
    required=False,
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def remove(
    ctx: click.core.Context,
    ids: str,
    verbose: bool,
    force: bool,
) -> None:
    """Remove records, ... from CoLRev repositories"""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )

    remove_operation = review_manager.get_remove_operation()

    if ids:
        remove_operation.remove_records(ids=ids)


@main.command(hidden=True, help_priority=30)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def docs(
    ctx: click.core.Context,
    verbose: bool,
    force: bool,
) -> None:
    """Show the CoLRev documentation."""

    webbrowser.open("https://colrev-environment.github.io/colrev/")


@main.command(help_priority=31)
@click.option(
    "--branch",
    help="Branch to merge.",
    required=False,
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def merge(
    ctx: click.core.Context,
    branch: str,
    verbose: bool,
    force: bool,
) -> None:
    """Merge git branches."""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )

    if not branch:
        colrev.ops.check.CheckOperation(review_manager)
        git_repo = review_manager.dataset.get_repo()
        print(f"possible branches: {','.join([b.name for b in git_repo.heads])}")
        return

    merge_operation = review_manager.get_merge_operation()
    merge_operation.main(branch=branch)


@main.command(help_priority=32)
@click.argument("selection")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose: printing more infos",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force mode",
)
@click.pass_context
def undo(
    ctx: click.core.Context,
    selection: str,
    verbose: bool,
    force: bool,
) -> None:
    """Undo operations."""

    review_manager = get_review_manager(
        ctx,
        {"verbose_mode": verbose, "force_mode": force},
    )

    if selection == "commit":
        colrev.ops.check.CheckOperation(review_manager)
        git_repo = review_manager.dataset.get_repo()
        git_repo.git.reset("--hard", "HEAD~1")


@main.command(help_priority=33)
@click.pass_context
def version(
    ctx: click.core.Context,
) -> None:
    """Show colrev version."""

    from importlib.metadata import version

    print(f'colrev version {version("colrev")}')


@main.command(help_priority=34)
@click.argument("packages", nargs=-1, required=False)
@click.option(
    "-U", "--upgrade", is_flag=True, help="Upgrade packages to latest version"
)
@click.option(
    "-e",
    "--editable",
    help="Install a project in editable mode from this path",
)
@click.option(
    "--force-reinstall",
    is_flag=True,
    help="Reinstall all packages even if they are already up-to-date",
)
@click.option(
    "--no-cache-dir",
    is_flag=True,
    help="Disable the cache",
)
def install(
    packages: typing.List[str],
    upgrade: bool,
    editable: str,
    force_reinstall: bool,
    no_cache_dir: bool,
) -> None:
    """Install packages

    To install all internal packages, run\n
        colrev install all_internal_packages"""

    if len(packages) == 1 and packages[0] == ".":
        review_manager = colrev.review_manager.ReviewManager()
        PACKAGE_MANAGER.install_project(
            review_manager=review_manager, force_reinstall=force_reinstall
        )

    else:

        PACKAGE_MANAGER.install(
            packages=packages,
            upgrade=upgrade,
            editable=editable,
            force_reinstall=force_reinstall,
            no_cache_dir=no_cache_dir,
        )



#! /usr/bin/env python
"""Exceptions of CoLRev."""
from __future__ import annotations

import typing
from pathlib import Path

from colrev.constants import Colors


class CoLRevException(Exception):
    """
    Base class for all exceptions raised by this package
    """


class RepoSetupError(CoLRevException):
    """
    The project files are not properly set up as a CoLRev project.
    """

    lr_docs = (
        "https://colrev-environment.github.io/colrev/manual/problem_formulation.html"
    )

    def __init__(self, msg: typing.Optional[str] = None) -> None:
        try:
            Path(".report.log").unlink(missing_ok=True)
        except PermissionError:
            pass
        if msg:
            self.message = f" {msg}"
        elif any(Path(Path.cwd()).iterdir()):
            self.message = (
                "Not an empty directory. "
                + "To create a CoLRev repository, navigate to an empty directory."
            )

        else:
            self.message = (
                "Not yet a CoLRev repository. "
                "To initialize a CoLRev repository (project), run:\n\n"
                f"   {Colors.ORANGE}colrev init --type {Colors.END}literature_review\n\n"
                f"Instead of {Colors.ORANGE}literature_review{Colors.END},"
                " you can use any of the following review type:\n\n"
                "  - narrative_review              : "
                "includes a manuscript\n"
                "  - scoping_review                : "
                "includes a manuscript, and a prisma diagram\n"
                "  - descriptive_review            : "
                "includes a manuscript, and a prisma diagram\n"
                "  - critical_review               : "
                "includes a manuscript, and a prisma diagram\n"
                "  - theoretical_review            : "
                "includes a manuscript\n"
                "  - conceptual_review             : "
                "includes a manuscript\n"
                "  - qualitative_systematic_review : "
                "includes a manuscript, data extraction tables, and a prisma diagram\n"
                "  - meta_analysis                 : "
                "includes a manuscript, data extraction tables, and a prisma diagram\n"
                "  - scientometric                 : "
                "includes a manuscript\n"
                "\nMore details about the differences between review types "
                f"are available at\n{self.lr_docs}"
            )

        super().__init__(self.message)


class CoLRevUpgradeError(CoLRevException):
    """
    The version of the local CoLRev package does not match with the CoLRev version
    used to create the latest commit in the project.
    An explicit upgrade of the data structures is needed.
    """

    def __init__(self, old: str, new: str) -> None:
        self.message = (
            f"Detected upgrade from {old} to {new}. To upgrade use\n     "
            f"{Colors.ORANGE}colrev upgrade{Colors.END}"
        )
        super().__init__(self.message)


class ReviewManagerNotNotifiedError(CoLRevException):
    """
    The ReviewManager was not notified about the operation.
    """

    def __init__(self) -> None:
        self.message = (
            "Create an operation and inform the review manager in advance"
            + " to avoid conflicts."
        )
        super().__init__(self.message)


class ParameterError(CoLRevException):
    """
    An invalid parameter was passed to CoLRev.
    """

    def __init__(self, *, parameter: str, value: str, options: list) -> None:
        options_string = "\n  - ".join(sorted(options))
        self.message = f"Invalid parameter {parameter}: {value}."
        if options:
            self.message += f"\n Options:\n  - {options_string}"
        super().__init__(self.message)


class InvalidSettingsError(CoLRevException):
    """
    Invalid value in SETTINGS_FILE.
    """

    def __init__(self, *, msg: str, fix_per_upgrade: bool = True) -> None:
        msg = f"Error in SETTINGS_FILE: {msg}"
        if fix_per_upgrade:
            msg += (
                "\nTo solve this, use\n  " f"{Colors.ORANGE}colrev upgrade{Colors.END}"
            )
        self.message = msg
        super().__init__(self.message)


# Valid commits, data structures, and repo states


class UnstagedGitChangesError(CoLRevException):
    """
    Unstaged git changes were found although a clean repository is required.
    """

    def __init__(self, changedFiles: list) -> None:
        self.message = (
            f"changes not yet staged: {changedFiles} (use git add . or stash)"
        )
        super().__init__(self.message)


class CleanRepoRequiredError(CoLRevException):
    """
    A clean git repository would be required.
    """

    def __init__(self, changedFiles: list, ignore_pattern: str) -> None:
        self.message = (
            "clean repository required (use git commit, discard or stash "
            + f"{changedFiles}; ignore_pattern={ignore_pattern})."
        )
        super().__init__(self.message)


class GitConflictError(CoLRevException):
    """
    There are git conflicts to be resolved before resuming operations.
    """

    def __init__(self, path: Path) -> None:
        self.message = f"please resolve git conflict in {path}"
        super().__init__(self.message)


class DirtyRepoAfterProcessingError(CoLRevException):
    """
    The git repository was not clean after completing the operation.
    """

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class GitNotAvailableError(CoLRevException):
    """
    Git is currently not available.
    """

    def __init__(self) -> None:
        self.message = "Git is currently not available (remove .git/index.lock exists)."
        super().__init__(self.message)


class AppendOnlyViolation(Exception):
    """Invalid changes to a file in append-only mode."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class ProcessOrderViolation(CoLRevException):
    """The process triggered dooes not have priority"""

    def __init__(
        self,
        operations_type: str,
        required_state: str,
        violating_records: list,
    ) -> None:
        self.message = (
            f" {operations_type}() requires all records to have at least "
            + f"'{required_state}', but there are records with {violating_records}."
        )
        super().__init__(self.message)


class StatusTransitionError(CoLRevException):
    """An invalid status transition was observed"""

    def __init__(self, msg: str) -> None:
        self.message = f" {msg}"
        super().__init__(self.message)


class NoRecordsError(CoLRevException):
    """The operation cannot be started because no records have been imported yet."""

    def __init__(self) -> None:
        self.message = "no records imported yet"
        super().__init__(self.message)


class FieldValueError(CoLRevException):
    """An error in field values was detected (in the main records)."""

    def __init__(self, msg: str) -> None:
        self.message = f" {msg}"
        super().__init__(self.message)


class MissingRecordQualityRuleSpecification(CoLRevException):
    """A quality rule is missing."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class StatusFieldValueError(CoLRevException):
    """An error in the status field values was detected."""

    def __init__(self, record: str, status_type: str, status_value: str) -> None:
        self.message = f"{status_type} set to '{status_value}' in {record}."
        super().__init__(self.message)


class OriginError(CoLRevException):
    """An error in the colrev_origin field values was detected."""

    def __init__(self, msg: str) -> None:
        self.message = f" {msg}"
        super().__init__(self.message)


class DuplicateIDsError(CoLRevException):
    """Duplicate IDs were detected."""

    def __init__(self, msg: str) -> None:
        self.message = f" {msg}"
        super().__init__(self.message)


class NotEnoughDataToIdentifyException(CoLRevException):
    """The meta-data is not sufficiently complete to identify the record."""

    def __init__(
        self,
        *,
        msg: typing.Optional[str] = None,
        missing_fields: typing.Optional[list] = None,
    ) -> None:
        self.message = msg
        self.missing_fields = missing_fields
        super().__init__(self.message)


class NotTOCIdentifiableException(CoLRevException):
    """The record cannot be identified through table-of-contents.
    Either the table-of-contents key is not implemented or
    the ENTRYTPE is not organized in tables-of-contents (e.g., online)."""

    def __init__(self, msg: typing.Optional[str] = None) -> None:
        self.message = msg
        super().__init__(self.message)


class RecordNotInTOCException(CoLRevException):
    """The record is not part of the table-of-contents (TOC)."""

    def __init__(self, *, record_id: str, toc_key: str) -> None:
        self.record_id = record_id
        self.toc_key = toc_key
        self.message = f"{record_id} not part of toc: {toc_key}"
        super().__init__(self.message)


class PropagatedIDChange(CoLRevException):
    """Changes in propagated records ID detected."""

    def __init__(self, notifications: list) -> None:
        self.message = "\n    ".join(notifications)
        super().__init__("Attempt to change propagated IDs:" + self.message)


# Init


class NonEmptyDirectoryError(CoLRevException):
    """Trying to initialize CoLRev in a non-empty directory."""

    def __init__(self, *, filepath: Path, content: list) -> None:
        if len(content) > 3:
            content = content[0:5] + ["..."]
        self.message = (
            f"Directory {filepath} not empty "
            + f"(files: {', '.join(content)}). "
            + "\nPlease change to an empty directory to initialize a project."
        )
        super().__init__(self.message)


class RepoInitError(CoLRevException):
    """Error during initialization of CoLRev project."""

    def __init__(self, *, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


# Search


class SearchNotAutomated(CoLRevException):
    """The search cannot be completed automatically."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class InvalidQueryException(CoLRevException):
    """The query format is not valid."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class RecordNotParsableException(CoLRevException):
    """The record could not be parsed."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class NotFeedIdentifiableException(CoLRevException):
    """The record does not contain the required source_identifier (cannot be added to the feed)."""

    def __init__(self, msg: typing.Optional[str] = None) -> None:
        self.message = msg
        super().__init__(self.message)


class SearchSourceException(CoLRevException):
    """Records cannot be retrieved from the SearchSource."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


class RecordNotFoundException(CoLRevException):
    """Record not found in the main records file."""

    def __init__(self, msg: str) -> None:
        self.message = msg
        super().__init__(self.message)


# Load


class ImportException(CoLRevException):
    """An error occured in the import functions."""

    def __init__(
        self,
        msg: str,
    ) -> None:
        self.message = msg
        super().__init__(self.message)


class NoRecordsToImport(CoLRevException):
    """No records to import."""

    def __init__(
        self,
        msg: str,
    ) -> None:
        self.message = msg
        super().__init__(self.message)


class UnsupportedImportFormatError(CoLRevException):
    """The file format is not supported."""

    def __init__(
        self,
        import_path: Path,
    ) -> None:
        self.import_path = import_path
        self.message = (
            f"Format of SearchSource file not supported ({self.import_path.name}) "
        )
        super().__init__(self.message)


# Prep


class RecordNotFoundInPrepSourceException(CoLRevException):
    """The record was not found in the prep search source."""

    def __init__(
        self,
        *,
        msg: str,
    ) -> None:
        self.message = msg
        super().__init__(self.message)


# Dedupe


class DedupeError(CoLRevException):
    """An exception in the dedupe operation"""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


# Data


class DataException(CoLRevException):
    """Exception in the data operation"""

    def __init__(self, *, msg: str) -> None:
        self.message = msg
        super().__init__("DataException: " + self.message)


# Push


class RecordNotInRepoException(CoLRevException):
    """The record was not found in the main records."""

    def __init__(self, record_id: str) -> None:
        if id is not None:
            self.message = f"Record not in repository ({record_id})"
        else:
            self.message = "Record not in repository"
        super().__init__(self.message)


class CorrectionPreconditionException(CoLRevException):
    """Precondition for corrections not given (clean git repository)."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


# PDF Hash service


class InvalidPDFException(CoLRevException):
    """The PDF is invalid (empty, encrypted or broken)."""

    def __init__(self, path: Path) -> None:
        self.message = f"Invalid PDF (empty/broken): {path}"
        super().__init__(self.message)


class PDFHashError(CoLRevException):
    """An error occurred during PDF hashing."""

    def __init__(self, path: Path) -> None:
        self.message = f"Error during PDF hashing: {path}"
        super().__init__(self.message)


# Environment services


class MissingDependencyError(CoLRevException):
    """The required dependency is not available."""

    def __init__(self, dep: str) -> None:
        self.message = f"{dep}"
        super().__init__(self.message)


class PackageParameterError(CoLRevException):
    """The parameter for the package are not correct."""

    def __init__(self, dep: str) -> None:
        self.message = f"{dep}"
        super().__init__(self.message)


class DependencyConfigurationError(CoLRevException):
    """The required dependency is not configured correctly."""

    def __init__(self, dep: str) -> None:
        self.message = f"{dep}"
        super().__init__(self.message)


class ServiceNotAvailableException(CoLRevException):
    """An environment service is not available."""

    def __init__(self, dep: str, detailed_trace: str = "") -> None:
        self.dep = dep
        self.detailed_trace = detailed_trace
        super().__init__(f"Service not available: {self.dep}")


class PortAlreadyRegisteredException(CoLRevException):
    """The port (localhost) is already registered."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class TEITimeoutException(CoLRevException):
    """A timeout occurred during TEI generation"""


class TEIException(CoLRevException):
    """An exception related to the TEI format"""


class RecordNotInIndexException(CoLRevException):
    """The requested record was not found in the LocalIndex."""

    def __init__(self, record_id: str) -> None:
        if id is not None:
            self.message = f"Record not in index ({record_id})"
        else:
            self.message = "Record not in index"
        super().__init__(self.message)


class RecordNotIndexableException(CoLRevException):
    """The requested record could not be added to the LocalIndex."""

    def __init__(
        self,
        record_id: typing.Optional[str] = None,
        missing_key: typing.Optional[str] = None,
    ) -> None:
        self.missing_key = missing_key
        if missing_key is None:
            missing_key = "-"
        if record_id is not None:
            self.message = f"Record cannot be indexed ({record_id}): {missing_key}"
        else:
            self.message = f"Record cannot be indexed: {missing_key}"
        super().__init__(self.message)


class TOCNotAvailableException(CoLRevException):
    """Tables of contents (toc) are not available for the requested item."""

    def __init__(self, msg: typing.Optional[str] = None) -> None:
        self.message = msg
        super().__init__(self.message)


class CuratedOutletNotUnique(CoLRevException):
    """The outlets (journals or conferences) with curated metadata are not unique."""

    def __init__(self, msg: typing.Optional[str] = None) -> None:
        self.message = msg
        super().__init__(self.message)


class InvalidLanguageCodeException(CoLRevException):
    """Language code field does not comply with the required standard."""

    def __init__(self, invalid_language_codes: list) -> None:
        self.invalid_language_codes = invalid_language_codes

        super().__init__(f"Invalid language codes: {', '.join(invalid_language_codes)}")


class PackageSettingMustStartWithPackagesException(CoLRevException):
    """package settings must start with `packages` key"""

    def __init__(self, invalid_key: str) -> None:
        self.invalid_key = invalid_key
        super().__init__(
            f"Package settings must start with `packages` key. {invalid_key}"
        )


class TemplateNotAvailableError(CoLRevException):
    """The requested template is not available."""

    def __init__(self, template_path: str) -> None:
        self.message = f"Template not available: {template_path}"
        super().__init__(self.message)



#! /usr/bin/env python
"""Functionality for individual records."""
from __future__ import annotations

import pprint
import typing
from copy import deepcopy
from pathlib import Path

import dictdiffer

import colrev.exceptions as colrev_exceptions
import colrev.record.record_identifier
import colrev.record.record_merger
import colrev.record.record_similarity
from colrev.constants import Colors
from colrev.constants import DefectCodes
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import RecordState

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager
    import colrev.record.qm.quality_model


# pylint: disable=too-many-public-methods


class Record:
    """The Record class provides a range of basic convenience functions"""

    pp = pprint.PrettyPrinter(indent=4, width=140, compact=False)

    def __init__(self, data: dict) -> None:
        self.data = data
        """Dictionary containing the record data"""
        # Note : avoid parsing upon Record instantiation as much as possible
        # to maintain high performance and ensure pickle-abiligy (in multiprocessing)

    def __repr__(self) -> str:  # pragma: no cover
        return self.pp.pformat(self.data)

    def __str__(self) -> str:
        ik_order = [Fields.ID, Fields.ENTRYTYPE]
        ik_order += [k for k in FieldSet.IDENTIFYING_FIELD_KEYS if k in self.data]
        ck_order = [k for k, v in self.data.items() if k not in ik_order]
        ik_sorted = {k: v for k, v in self.data.items() if k in ik_order}
        ck_sorted = {k: v for k, v in self.data.items() if k in ck_order}
        ret_str = self.pp.pformat(ik_sorted)[:-1] + "\n"
        ret_str += self.pp.pformat(ck_sorted)[1:]
        return ret_str

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__

    def get_citation_format(self) -> str:
        """Get the record as a citation"""
        formatted_ref = (
            f"{self.data.get(Fields.AUTHOR, '')} ({self.data.get(Fields.YEAR, '')}) "
            f"{self.data.get(Fields.TITLE, '')}. "
            f"{self.data.get(Fields.JOURNAL, '')}{self.data.get(Fields.BOOKTITLE, '')}, "
            f"{self.data.get(Fields.VOLUME, '')} ({self.data.get(Fields.NUMBER, '')})"
        )
        return formatted_ref

    def print_citation_format(self) -> None:
        """Print the record as a citation"""
        print(self.get_citation_format())

    def copy(self) -> Record:
        """Copy the record object"""
        return Record(deepcopy(self.data))

    def copy_prep_rec(self) -> colrev.record.record_prep.PrepRecord:
        """Copy the record object (as a PrepRecord)"""
        return colrev.record.record_prep.PrepRecord(deepcopy(self.data))

    def update_by_record(self, update_record: Record) -> None:
        """Update all data of a record object based on another record"""
        self.data = update_record.copy_prep_rec().get_data()

    def format_bib_style(self) -> str:
        """Simple formatter for bibliography-style output"""
        return (
            f"{self.data.get(Fields.AUTHOR, '')} "
            f"({self.data.get(Fields.YEAR, '')}) "
            f"{self.data.get(Fields.TITLE, '')}. "
            f"{self.data.get(Fields.JOURNAL, '')}"
            f"{self.data.get(Fields.BOOKTITLE, '')}, "
            f"({self.data.get(Fields.VOLUME, '')}) "
            f"{self.data.get(Fields.NUMBER, '')}"
        )

    def get_data(self) -> dict:
        """Get the record data"""

        if not isinstance(self.data.get(Fields.ORIGIN, []), list):
            self.data[Fields.ORIGIN] = self.data[Fields.ORIGIN].rstrip(";").split(";")
        assert isinstance(self.data.get(Fields.ORIGIN, []), list)

        return self.data

    def get_value(self, key: str, *, default: typing.Optional[str] = None) -> str:
        """Get a record value (based on the key parameter)"""
        if default is not None:
            try:
                ret = self.data[key]
                return ret
            except KeyError:
                return default
        else:
            return self.data[key]

    # pylint: disable=too-many-arguments
    def update_field(
        self,
        *,
        key: str,
        value: str,
        source: str,
        note: str = "",
        keep_source_if_equal: bool = True,
        append_edit: bool = True,
    ) -> None:
        """Update a record field (including provenance information)"""
        if keep_source_if_equal:
            if key in self.data:
                if self.data[key] == value:
                    return

        if key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if not self.masterdata_is_curated():
                if append_edit and key in self.data:
                    if key in self.data.get(Fields.MD_PROV, {}):
                        source = self.data[Fields.MD_PROV][key]["source"] + "|" + source
                    else:
                        source = "original|" + source
                self.add_field_provenance(key=key, source=source, note=note)
        else:
            if append_edit and key in self.data:
                if key in self.data.get(Fields.D_PROV, {}):
                    source = self.data[Fields.D_PROV][key]["source"] + "|" + source
                else:
                    source = "original|" + source
            self.add_field_provenance(key=key, source=source, note=note)
        self.data[key] = value

    def rename_field(self, *, key: str, new_key: str) -> None:
        """Rename a field"""
        if key not in self.data:
            return
        value = self.data[key]
        self.data[new_key] = value

        self.require_prov()
        if key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if key in self.data[Fields.MD_PROV]:
                value_provenance = self.data[Fields.MD_PROV][key]
                if "source" in value_provenance:
                    value_provenance["source"] += f"|rename-from:{key}"
            else:
                value_provenance = {
                    "source": f"|rename-from:{key}",
                    "note": "",
                }
            self.data[Fields.MD_PROV][new_key] = value_provenance
        else:
            if key in self.data[Fields.D_PROV]:
                value_provenance = self.data[Fields.D_PROV][key]
                if "source" in value_provenance:
                    value_provenance["source"] += f"|rename-from:{key}"
            else:
                value_provenance = {"source": f"|rename-from:{key}", "note": ""}
            self.data[Fields.D_PROV][new_key] = value_provenance

        self.remove_field(key=key)

    def remove_field(
        self, *, key: str, not_missing_note: bool = False, source: str = ""
    ) -> None:
        """Remove a field"""

        if key in self.data:
            del self.data[key]

        self.require_prov()
        if not_missing_note and key in FieldSet.MASTERDATA:
            # Example: journal without number
            # we should keep that information that a particular masterdata
            # field is not required
            if key not in self.data[Fields.MD_PROV]:
                self.data[Fields.MD_PROV][key] = {"source": "manual", "note": ""}
            self.data[Fields.MD_PROV][key]["note"] = f"IGNORE:{DefectCodes.MISSING}"
            if source != "":
                self.data[Fields.MD_PROV][key]["source"] = source
        else:
            if key in self.data.get(Fields.MD_PROV, {}):
                del self.data[Fields.MD_PROV][key]
            if key in self.data.get(Fields.D_PROV, {}):
                del self.data[Fields.D_PROV][key]

    def require_prov(self) -> None:
        """Ensure that provenance fields are available"""
        if Fields.MD_PROV not in self.data:
            self.data[Fields.MD_PROV] = {}
        if Fields.D_PROV not in self.data:
            self.data[Fields.D_PROV] = {}

    def align_provenance(self) -> None:
        """Remove unnecessary provenance information and add missing provenance information"""
        self.require_prov()
        for key in list(self.data[Fields.MD_PROV].keys()):
            if (
                key not in self.data
                and key != FieldValues.CURATED
                and "IGNORE:missing" not in self.get_field_provenance_notes(key)
            ):
                del self.data[Fields.MD_PROV][key]
        for key in list(self.data[Fields.D_PROV].keys()):
            if (
                key not in self.data
                and "IGNORE:missing" not in self.get_field_provenance_notes(key)
            ):
                del self.data[Fields.D_PROV][key]

        for non_prov_key in [Fields.ENTRYTYPE, Fields.ID] + FieldSet.PROVENANCE_KEYS:
            self.data[Fields.MD_PROV].pop(non_prov_key, None)
            self.data[Fields.D_PROV].pop(non_prov_key, None)

        for key in self.data.keys():
            if key in FieldSet.PROVENANCE_KEYS + [Fields.ID, Fields.ENTRYTYPE]:
                continue
            if key in FieldSet.IDENTIFYING_FIELD_KEYS:
                if (
                    not self.masterdata_is_curated()
                    and key not in self.data[Fields.MD_PROV]
                ):
                    self.data[Fields.MD_PROV][key] = {"source": "manual", "note": ""}
            elif key not in self.data[Fields.D_PROV]:
                self.data[Fields.D_PROV][key] = {"source": "manual", "note": ""}

    def add_provenance_all(self, *, source: str) -> None:
        """Add a data provenance (source) to all fields"""
        self.require_prov()
        for key in self.data.keys():
            if key in FieldSet.NO_PROVENANCE:
                continue
            if key in FieldSet.MASTERDATA:
                if self.masterdata_is_curated():
                    continue
                self.data[Fields.MD_PROV][key] = {"source": source, "note": ""}
            else:
                self.data[Fields.D_PROV][key] = {"source": source, "note": ""}

    def add_field_provenance(self, *, key: str, source: str, note: str = "") -> None:
        """Add a field provenance, including source and note (based on a key)"""

        if key in FieldSet.NO_PROVENANCE:
            return
        self.require_prov()

        if key in FieldSet.MASTERDATA:
            prov_dict = self.data[Fields.MD_PROV]
        else:
            prov_dict = self.data[Fields.D_PROV]

        if key not in prov_dict:
            prov_dict[key] = {"source": source, "note": note}
            return

        prov_dict[key]["source"] = source

        if prov_dict[key]["note"] == "" or note == "":
            prov_dict[key]["note"] = note
            return

        notes = prov_dict[key]["note"].split(",")
        if f"IGNORE:{note}" in notes:
            notes.remove(f"IGNORE:{note}")
            notes.append(note)
            prov_dict[key]["note"] = ",".join(notes)
        elif note in notes:
            return  # already added
        else:
            prov_dict[key]["note"] += f",{note}"

    def add_field_provenance_note(self, *, key: str, note: str) -> None:
        """Add a field provenance note (based on a key)"""
        if key in FieldSet.NO_PROVENANCE:
            return
        self.require_prov()

        if key in FieldSet.MASTERDATA:
            prov = self.data[Fields.MD_PROV]
        else:
            prov = self.data[Fields.D_PROV]

        if key not in prov:
            prov[key] = {
                "source": "ORIGINAL",
                "note": note,
            }
            return

        if prov[key]["note"] == "":
            prov[key]["note"] = note
        elif note not in prov[key]["note"].split(","):
            prov[key]["note"] += f",{note}"

    def get_field_provenance(
        self, *, key: str, default_source: str = "ORIGINAL"
    ) -> dict:
        """Get the provenance for a selected field (key)"""
        default_note = ""
        note = default_note
        source = default_source
        if key in FieldSet.MASTERDATA:
            if Fields.MD_PROV in self.data and key in self.data[Fields.MD_PROV]:
                if "source" in self.data[Fields.MD_PROV][key]:
                    source = self.data[Fields.MD_PROV][key]["source"]
                if "note" in self.data[Fields.MD_PROV][key]:
                    note = self.data[Fields.MD_PROV][key]["note"]
        else:
            if Fields.D_PROV in self.data and key in self.data[Fields.D_PROV]:
                if "source" in self.data[Fields.D_PROV][key]:
                    source = self.data[Fields.D_PROV][key]["source"]
                if "note" in self.data[Fields.D_PROV][key]:
                    note = self.data[Fields.D_PROV][key]["note"]

        return {"source": source, "note": note}

    def get_field_provenance_notes(self, key: str) -> list:
        """Get field provenance notes based on a key"""
        if key in FieldSet.MASTERDATA:
            if Fields.MD_PROV not in self.data:
                return []
            if key not in self.data[Fields.MD_PROV]:
                return []
            notes = self.data[Fields.MD_PROV][key]["note"].split(",")
        else:
            if Fields.D_PROV not in self.data:
                return []
            if key not in self.data[Fields.D_PROV]:
                return []
            notes = self.data[Fields.D_PROV][key]["note"].split(",")
        return [note.strip() for note in notes if note]

    def get_field_provenance_source(self, key: str) -> str:
        """Get the provenance source for a selected field (key)"""
        return self.get_field_provenance(key=key)["source"]

    def remove_field_provenance_note(self, *, key: str, note: str) -> None:
        """Remove field provenance notes based on a key (also if IGNORE:note)"""
        if key in FieldSet.MASTERDATA:
            if Fields.MD_PROV not in self.data:
                return
            if key not in self.data[Fields.MD_PROV]:
                return
            notes = self.data[Fields.MD_PROV][key]["note"].split(",")
            if note not in notes:
                return
            self.data[Fields.MD_PROV][key]["note"] = ",".join(
                n for n in notes if n not in [note, f"IGNORE:{note}"]
            )

        else:
            if Fields.D_PROV not in self.data:
                return
            if key not in self.data[Fields.D_PROV]:
                return
            notes = self.data[Fields.D_PROV][key]["note"].split(",")
            if note not in notes:
                return
            self.data[Fields.D_PROV][key]["note"] = ",".join(
                n for n in notes if n not in [note, f"IGNORE:{note}"]
            )

    def complete_provenance(self, *, source_info: str) -> bool:
        """Complete provenance information for indexing"""

        for key in list(self.data.keys()):
            if key in FieldSet.NO_PROVENANCE:
                continue

            if key in FieldSet.MASTERDATA:
                if not self.masterdata_is_curated():
                    self.add_field_provenance(key=key, source=source_info, note="")
            else:
                self.add_field_provenance(key=key, source=source_info, note="")

        return True

    def set_masterdata_complete(
        self, *, source: str, masterdata_repository: bool, replace_source: bool = True
    ) -> None:
        """Set the masterdata to complete"""
        # pylint: disable=too-many-branches
        if self.masterdata_is_curated() or masterdata_repository:
            return

        self.require_prov()
        md_p_dict = self.data[Fields.MD_PROV]

        for identifying_field_key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if identifying_field_key in [Fields.AUTHOR, Fields.TITLE, Fields.YEAR]:
                continue
            if self.data.get(identifying_field_key, "NA") == FieldValues.UNKNOWN:
                del self.data[identifying_field_key]
            if identifying_field_key in md_p_dict:
                note = md_p_dict[identifying_field_key]["note"]
                if (
                    DefectCodes.MISSING in note
                    and f"IGNORE:{DefectCodes.MISSING}" not in note
                ):
                    md_p_dict[identifying_field_key]["note"] = note.replace(
                        DefectCodes.MISSING, ""
                    )

        if self.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            if Fields.VOLUME not in self.data:
                if Fields.VOLUME in self.data[Fields.MD_PROV]:
                    self.data[Fields.MD_PROV][Fields.VOLUME][
                        "note"
                    ] = f"IGNORE:{DefectCodes.MISSING}"
                    if replace_source:
                        self.data[Fields.MD_PROV][Fields.VOLUME]["source"] = source
                else:
                    self.data[Fields.MD_PROV][Fields.VOLUME] = {
                        "source": source,
                        "note": f"IGNORE:{DefectCodes.MISSING}",
                    }

            if Fields.NUMBER not in self.data:
                if Fields.NUMBER in self.data[Fields.MD_PROV]:
                    self.data[Fields.MD_PROV][Fields.NUMBER][
                        "note"
                    ] = f"IGNORE:{DefectCodes.MISSING}"
                    if replace_source:
                        self.data[Fields.MD_PROV][Fields.NUMBER]["source"] = source
                else:
                    self.data[Fields.MD_PROV][Fields.NUMBER] = {
                        "source": source,
                        "note": f"IGNORE:{DefectCodes.MISSING}",
                    }

    def set_masterdata_consistent(self) -> None:
        """Set the masterdata to consistent"""
        self.require_prov()
        md_p_dict = self.data[Fields.MD_PROV]

        for identifying_field_key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if identifying_field_key in md_p_dict:
                note = md_p_dict[identifying_field_key]["note"]
                if DefectCodes.INCONSISTENT_WITH_ENTRYTYPE in note:
                    md_p_dict[identifying_field_key]["note"] = note.replace(
                        DefectCodes.INCONSISTENT_WITH_ENTRYTYPE, ""
                    )

    def reset_pdf_provenance_notes(self) -> None:
        """Reset the PDF (file) provenance notes"""
        if Fields.D_PROV not in self.data:
            self.add_field_provenance_note(key=Fields.FILE, note="")
        else:
            if Fields.FILE in self.data[Fields.D_PROV]:
                self.data[Fields.D_PROV][Fields.FILE]["note"] = ""
            else:
                self.data[Fields.D_PROV][Fields.FILE] = {
                    "source": "NA",
                    "note": "",
                }

    def defects(self, key: str) -> typing.List[str]:
        """Get a list of defects for a field"""
        self.require_prov()
        defects = []
        if key in self.data[Fields.MD_PROV]:
            defects.extend(self.data[Fields.MD_PROV][key]["note"].split(","))
        if key in self.data[Fields.D_PROV]:
            defects.extend(self.data[Fields.D_PROV][key]["note"].split(","))
        return defects

    def has_quality_defects(self, *, key: str = "") -> bool:
        """Check whether a record (or specific field/key) has quality defects"""
        if key != "":
            if key in self.data.get(Fields.MD_PROV, {}):
                note = self.data[Fields.MD_PROV][key]["note"]
                notes = note.split(",")
                notes = [n for n in notes if not n.startswith("IGNORE:")]
                return len(notes) != 0
            if key in self.data.get(Fields.D_PROV, {}):
                note = self.data[Fields.D_PROV][key]["note"]
                notes = note.split(",")
                notes = [n for n in notes if not n.startswith("IGNORE:")]
                return len(notes) != 0
            return False
        defect_codes = [
            n
            for x in self.data.get(Fields.MD_PROV, {}).values()
            for n in x["note"].split(",")
            if not n.startswith("IGNORE:") and n != ""
        ]
        return bool(defect_codes)

    def has_fatal_quality_defects(self) -> bool:
        """Check whether a record has fatal quality defects"""

        required_fields = [Fields.TITLE, Fields.AUTHOR, Fields.YEAR]
        if not all(r in self.data for r in required_fields) or not all(
            self.data[r] != FieldValues.UNKNOWN for r in required_fields
        ):
            return True

        if (
            self.data.get(Fields.JOURNAL, FieldValues.UNKNOWN) == FieldValues.UNKNOWN
            and self.data.get(Fields.BOOKTITLE, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
            and self.data[Fields.ENTRYTYPE]
            in [ENTRYTYPES.ARTICLE, ENTRYTYPES.INPROCEEDINGS]
        ):
            return True

        if (
            DefectCodes.IDENTICAL_VALUES_BETWEEN_TITLE_AND_CONTAINER
            in self.get_field_provenance_notes(Fields.TITLE)
        ):
            return True

        # title if it starts with "doi:"
        if Fields.TITLE in self.data:
            if self.data[Fields.TITLE].lower().startswith("doi:"):
                return True

        # title has more numbers than characters
        if Fields.TITLE in self.data:
            if sum(c.isdigit() for c in self.data[Fields.TITLE]) > sum(
                c.isalpha() for c in self.data[Fields.TITLE]
            ):
                return True

        return False

    def has_pdf_defects(self) -> bool:
        """Check whether the PDF has quality defects"""

        return bool(
            [
                n
                for n in self.get_field_provenance_notes(Fields.FILE)
                if not n.startswith("IGNORE:") and n != ""
            ]
        )

    def ignore_defect(self, *, key: str, defect: str) -> None:
        """Ignore a defect for a field"""

        ignore_code = f"IGNORE:{defect}"
        self.remove_field_provenance_note(key=key, note=defect)
        self.add_field_provenance_note(key=key, note=ignore_code)

    def ignored_defect(self, *, key: str, defect: str) -> bool:
        """Get a list of ignored defects for a record"""
        ignore_code = f"IGNORE:{defect}"
        if Fields.MD_PROV in self.data and key in self.data[Fields.MD_PROV]:
            notes = self.data[Fields.MD_PROV][key]["note"].split(",")
            return ignore_code in notes
        if Fields.D_PROV in self.data and key in self.data[Fields.D_PROV]:
            notes = self.data[Fields.D_PROV][key]["note"].split(",")
            return ignore_code in notes
        return False

    def get_container_title(self, *, na_string: str = "NA") -> str:
        """Get the record's container title (journal name, booktitle, etc.)"""

        if Fields.ENTRYTYPE not in self.data:
            return self.data.get(
                Fields.JOURNAL, self.data.get(Fields.BOOKTITLE, na_string)
            )
        if self.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            return self.data.get(Fields.JOURNAL, na_string)
        if self.data[Fields.ENTRYTYPE] in [
            ENTRYTYPES.INPROCEEDINGS,
            ENTRYTYPES.PROCEEDINGS,
            ENTRYTYPES.INBOOK,
        ]:
            return self.data.get(Fields.BOOKTITLE, na_string)
        if self.data[Fields.ENTRYTYPE] == ENTRYTYPES.BOOK:
            return self.data.get(Fields.TITLE, na_string)
        return na_string

    def set_masterdata_curated(self, source: str) -> None:
        """Set record masterdata to curated"""
        self.data[Fields.MD_PROV] = {
            FieldValues.CURATED: {"source": source, "note": ""}
        }

    def masterdata_is_curated(self) -> bool:
        """Check whether the record masterdata is curated"""
        return FieldValues.CURATED in self.data.get(Fields.MD_PROV, {})

    def get_colrev_id(
        self,
        *,
        assume_complete: bool = False,
    ) -> str:
        """Returns the colrev_id of the Record."""

        return colrev.record.record_identifier.get_colrev_id(
            self,
            assume_complete=assume_complete,
        )

    @classmethod
    def get_colrev_pdf_id(
        cls,
        pdf_path: Path,
    ) -> str:  # pragma: no cover
        """Generate the colrev_pdf_id"""

        return colrev.record.record_identifier.get_colrev_pdf_id(pdf_path)

    def get_toc_key(self) -> str:
        """Get the record's toc-key"""
        return colrev.record.record_identifier.get_toc_key(self)

    def prescreen_exclude(self, *, reason: str, print_warning: bool = False) -> None:
        """Prescreen-exclude a record"""
        # Warn when setting rev_synthesized/rev_included to prescreen_excluded
        # Especially in cases in which the prescreen-exclusion decision
        # is revised (e.g., because a paper was retracted)
        # In these cases, the paper may already be in the data extraction/synthesis
        if self.data.get(Fields.STATUS, "NA") in [
            RecordState.rev_synthesized,
            RecordState.rev_included,
        ]:
            print(
                f"\n{Colors.RED}Warning: setting paper to prescreen_excluded. Please check and "
                f"remove from synthesis: {self.data['ID']}{Colors.END}\n"
            )

        self.set_status(RecordState.rev_prescreen_excluded)

        if (
            FieldValues.RETRACTED not in self.data.get(Fields.PRESCREEN_EXCLUSION, "")
            and FieldValues.RETRACTED == reason
        ):
            if print_warning:
                print(
                    f"\n{Colors.RED}Paper retracted and prescreen "
                    f"excluded: {self.data['ID']}{Colors.END}\n"
                )
        if FieldValues.RETRACTED == reason:
            self.data[Fields.RETRACTED] = FieldValues.RETRACTED

        self.data[Fields.PRESCREEN_EXCLUSION] = reason

        # Note: when records are prescreen-excluded during prep:
        to_drop = []
        for key, value in self.data.items():
            if value == FieldValues.UNKNOWN:
                to_drop.append(key)
        for key in to_drop:
            self.remove_field(key=key)

    def get_tei_filename(self) -> Path:
        """Get the TEI filename associated with the file (PDF)"""
        tei_filename = Path(f".tei/{self.data[Fields.ID]}.tei.xml")
        if Fields.FILE in self.data:
            tei_filename = Path(
                self.data[Fields.FILE].replace("pdfs/", ".tei/")
            ).with_suffix(".tei.xml")
        return tei_filename

    def run_pdf_quality_model(
        self,
        pdf_qm: colrev.record.qm.quality_model.QualityModel,
        *,
        set_prepared: bool = False,
    ) -> None:
        """Run the PDF quality model"""

        pdf_qm.run(record=self)
        if self.has_pdf_defects():
            self.set_status(RecordState.pdf_needs_manual_preparation)
        elif set_prepared:
            self.set_status(RecordState.pdf_prepared)

    def run_quality_model(
        self,
        quality_model: colrev.record.qm.quality_model.QualityModel,
        *,
        set_prepared: bool = False,
    ) -> None:
        """Update the masterdata provenance"""

        self.require_prov()
        self.is_retracted()

        if self.masterdata_is_curated() and set_prepared:
            self.set_status(RecordState.md_prepared)
            return

        # Apply the checkers (including field key requirements etc.)
        quality_model.run(record=self)

        if not set_prepared:
            return
        if (
            Fields.STATUS in self.data
            and self.data[Fields.STATUS] == RecordState.rev_prescreen_excluded
        ):
            return

        if self.has_fatal_quality_defects():
            self.set_status(RecordState.md_needs_manual_preparation)
        else:
            self.set_status(RecordState.md_prepared)

    def is_retracted(self) -> bool:
        """Check for potential retracts"""

        # Legacy
        if (
            self.data.get("crossmark", "") == "True"
            or self.data.get("colrev.crossref.crossmark", "") == "True"
        ):
            self.prescreen_exclude(reason=FieldValues.RETRACTED, print_warning=True)
            self.remove_field(key="crossmark")
            self.remove_field(key="colrev.crossref.crossmark")
        if self.data.get("warning", "") == "Withdrawn (according to DBLP)":
            self.prescreen_exclude(reason=FieldValues.RETRACTED, print_warning=True)
            self.remove_field(key="warning")

        if Fields.RETRACTED in self.data:
            self.prescreen_exclude(reason=FieldValues.RETRACTED, print_warning=True)
            return True

        return False

    # pylint: disable=too-many-branches
    def change_entrytype(
        self,
        new_entrytype: str,
        *,
        qm: colrev.record.qm.quality_model.QualityModel,
    ) -> None:
        """Change the ENTRYTYPE"""
        if new_entrytype == self.data.get(Fields.ENTRYTYPE, "NA"):
            if Fields.MD_PROV in self.data:
                self.align_provenance()
            return  # otherwise, IGNORE:missing would be reset
        for value in self.data.get(Fields.MD_PROV, {}).values():
            if any(
                x in value["note"]
                for x in [DefectCodes.INCONSISTENT_WITH_ENTRYTYPE, DefectCodes.MISSING]
            ):
                value["note"] = ""
        missing_fields = [k for k, v in self.data.items() if v == FieldValues.UNKNOWN]
        for missing_field in missing_fields:
            self.remove_field(key=missing_field)

        self.align_provenance()

        self.data[Fields.ENTRYTYPE] = new_entrytype
        if new_entrytype in [ENTRYTYPES.INPROCEEDINGS, ENTRYTYPES.PROCEEDINGS]:
            if Fields.JOURNAL in self.data and Fields.BOOKTITLE not in self.data:
                self.rename_field(key=Fields.JOURNAL, new_key=Fields.BOOKTITLE)
        elif new_entrytype == ENTRYTYPES.ARTICLE:
            if Fields.BOOKTITLE in self.data:
                self.rename_field(key=Fields.BOOKTITLE, new_key=Fields.JOURNAL)
        elif new_entrytype in [
            ENTRYTYPES.INBOOK,
            ENTRYTYPES.BOOK,
            ENTRYTYPES.INCOLLECTION,
            ENTRYTYPES.PHDTHESIS,
            ENTRYTYPES.THESIS,
            ENTRYTYPES.MASTERSTHESIS,
            ENTRYTYPES.BACHELORTHESIS,
            ENTRYTYPES.TECHREPORT,
            ENTRYTYPES.UNPUBLISHED,
            ENTRYTYPES.MISC,
            ENTRYTYPES.SOFTWARE,
            ENTRYTYPES.ONLINE,
            ENTRYTYPES.CONFERENCE,
        ]:
            pass
        else:
            raise colrev_exceptions.MissingRecordQualityRuleSpecification(
                f"No ENTRYTYPE specification ({new_entrytype})"
            )

        self.run_quality_model(qm, set_prepared=True)

    def set_status(self, target_state: RecordState, *, force: bool = False) -> None:
        """Set the record status"""

        if RecordState.md_prepared == target_state and not force:
            if self.has_fatal_quality_defects():
                target_state = RecordState.md_needs_manual_preparation
        # pylint: disable=colrev-direct-status-assign
        self.data[Fields.STATUS] = target_state

    def get_diff(
        self, other_record: Record, *, identifying_fields_only: bool = True
    ) -> list:
        """Get diff between record objects"""

        if not identifying_fields_only:
            return list(dictdiffer.diff(self.get_data(), other_record.get_data()))

        diff = []
        for selected_tuple in list(
            dictdiffer.diff(self.get_data(), other_record.get_data())
        ):
            if selected_tuple[0] == "change":
                if selected_tuple[1] in FieldSet.IDENTIFYING_FIELD_KEYS:
                    diff.append(selected_tuple)
            if selected_tuple[0] == "add":
                addition_list: typing.Tuple = ("add", "", [])
                for addition_item in selected_tuple[2]:
                    if addition_item[0] in FieldSet.IDENTIFYING_FIELD_KEYS:
                        addition_list[2].append(addition_item)
                if addition_list[2]:
                    diff.append(addition_list)
            if selected_tuple[0] == "remove":
                removal_list: typing.Tuple = ("remove", "", [])
                for removal_item in selected_tuple[2]:
                    if removal_item[0] in FieldSet.IDENTIFYING_FIELD_KEYS:
                        removal_list[2].append(removal_item)
                if removal_list[2]:
                    diff.append(removal_list)
        return diff

    @classmethod
    def get_record_change_score(cls, record_a: Record, record_b: Record) -> float:
        """Determine how much records changed

        This method is less sensitive than get_record_similarity, especially when
        fields are missing. For example, if the journal field is missing in both
        records, get_similarity will return a value > 1.0. The get_record_changes
        will return 0.0 (if all other fields are equal)."""

        return colrev.record.record_similarity.get_record_change_score(
            record_a, record_b
        )

    @classmethod
    def get_record_similarity(cls, record_a: Record, record_b: Record) -> float:
        """Determine the similarity between two records (their masterdata)"""

        return colrev.record.record_similarity.get_record_similarity(record_a, record_b)

    def merge(
        self,
        merging_record: Record,
        *,
        default_source: str,
        preferred_masterdata_source_prefixes: typing.Optional[list] = None,
    ) -> None:
        """General-purpose record merging
        for preparation, curated/non-curated records and records with origins

        Apply heuristics to create a fusion of the best fields based on
        quality heuristics"""

        if preferred_masterdata_source_prefixes is None:
            preferred_masterdata_source_prefixes = []

        colrev.record.record_merger.merge(
            self,
            merging_record,
            default_source=default_source,
            preferred_masterdata_source_prefixes=preferred_masterdata_source_prefixes,
        )



#! /usr/bin/env python
"""GROBID service to extract and annotate PDF contents."""
from __future__ import annotations

import logging
import time
import typing

import docker
import requests

import colrev.env.docker_manager
import colrev.env.environment_manager


class GrobidService:
    """An environment service for machine readability/annotation (PDF to TEI conversion)"""

    GROBID_URL = "http://localhost:8070"
    GROBID_IMAGE = "lfoppiano/grobid:0.8.1"

    def __init__(
        self,
        *,
        environment_manager: typing.Optional[
            colrev.env.environment_manager.EnvironmentManager
        ] = None,
    ) -> None:
        colrev.env.docker_manager.DockerManager.build_docker_image(
            imagename=self.GROBID_IMAGE
        )
        self.start(environment_manager)
        self.check_grobid_availability()

    def check_grobid_availability(self, *, wait: bool = True) -> bool:
        """Check whether the GROBID service is available"""
        i = 0
        while True:
            i += 1
            time.sleep(1)
            try:
                ret = requests.get(self.GROBID_URL + "/api/isalive", timeout=30)
                if ret.text == "true":
                    return True
            except requests.exceptions.ConnectionError:
                pass
            if not wait:
                return False
            if i == -1:
                break
            if i > 20:
                raise requests.exceptions.ConnectionError()
        return True

    def start(
        self,
        environment_manager: typing.Optional[
            colrev.env.environment_manager.EnvironmentManager
        ] = None,
    ) -> None:
        """Start the GROBID service"""
        # pylint: disable=consider-using-with

        try:
            res = self.check_grobid_availability(wait=False)
            if res:
                return
        except requests.exceptions.ConnectionError:
            pass

        client = docker.from_env()
        logging.info("Running docker container created from %s", self.GROBID_IMAGE)
        logging.info("Starting grobid service...")
        client.containers.run(
            self.GROBID_IMAGE,
            auto_remove=True,
            tty=True,
            mem_limit="4g",
            ports={8070: 8070, 8071: 8071},
            detach=True,
        )
        if environment_manager:
            environment_manager.register_ports(["8070", "8071"])

        self.check_grobid_availability()
